#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <glib.h>

#include "config.h"
#include "logger.h"
#include "mcplay.h"
#include "custom.h"

/* helper func */
gchar** strsplit( const char *string )
{
   gchar *tok, *c0, *c1;
   gchar **res = NULL;
   gint i=0;
   
   tok = g_strdup(string);
   
   c0 = tok;
   
   while(TRUE)
   {
      i++;
      c0 = g_strstrip(c0);
      if( *c0 == '\'' )
      {  c0++; c1 = strchr(c0, '\''); }
      else if( *c0 == '"' )
      {  c0++; c1 = strchr(c0, '"'); }
      else
         c1 = strchr(c0, ' ');
         
      if( c1 == NULL )
      {  res = g_renew(gchar*, res, i+1);
         res[i-1] = g_strdup(c0);
         break;
      }
      
      *c1 = '\0';
      res = g_renew(gchar*, res, i+1);
      res[i-1] = g_strdup(c0);
      c0 = ++c1;
   }
   res[i] = NULL;
   
   return res;
}

typedef struct _CustomData
{
   gchar **command;
   gint pid;
   GimmeNextFunc gimme_next;
   gpointer next_data;
   gboolean stopforced;
   gint fdnull;
} CustomData;

static CustomData *sigdata = NULL;
static GMainLoop *loop = NULL;

void custom_sigchld( int signal )
{
   int status, r;
   
   g_debug(__FILE__": got SIGCHLD ... ");
   r=waitpid(-1, &status, WNOHANG);
   
   if( r == 0 )
   {  g_debug("just paused?");
      return;
   }
   else if( r != sigdata->pid )
   {  g_debug("pid %d, not this player's child?", r);
      return;
   }
   
   sigdata->pid = -1;
   
   if( sigdata->gimme_next != NULL && !sigdata->stopforced )
      g_timeout_add( 1, sigdata->gimme_next, sigdata->next_data);
      
   if( g_main_is_running(loop) )
      g_main_quit(loop);
   
   g_debug("ok. (pid %d, stat %d)", r, WEXITSTATUS(status));
}

gboolean custom_kill_handler( CustomData *data )
{
   gint r;
   
   kill(-1*data->pid, SIGCONT);
   r=kill(-1*data->pid, SIGKILL);
   
   if( r < 0 )
      g_info(__FILE__": cannot kill: %s", g_strerror(errno));
   
   return FALSE;
}

gint custom_stop( CustomData *data )
{
   g_debug(__FILE__": custom_stop()");
   
   if( data->pid > 0 )
   {
      data->stopforced = TRUE;
      
      /* we wait until the signal arrives that our child is dead :-( */
      g_timeout_add( 1, (GSourceFunc)custom_kill_handler, data);
      g_main_run(loop);
      g_debug(__FILE__": player stopped.");
   }
   
   return 0;
}

gint custom_play( CustomData *data, const char *file )
{
   /* char buf[PATH_MAX+NAME_MAX]; */
   char **argv = NULL;
   int i;
   
   if( data->pid > 0 )
   {  data->stopforced = TRUE;
      custom_stop(data);
   }
   
   data->stopforced = FALSE;
   sigdata = data;   
   signal(SIGCHLD, custom_sigchld);

   /* g_snprintf(buf, sizeof(buf)-1, data->command, file );
   argv = strsplit(buf); */
   
   i=0;
   while( data->command[i] != NULL )
   {
      argv = g_renew(char*, argv, i+2);
      if( strcmp(data->command[i], "%s") == 0 )
         argv[i] = g_strdup(file);
      else
         argv[i] = g_strdup(data->command[i]);
      i++;
   }
   argv[i] = NULL;
   
   if( (data->pid=fork()) == 0 )
   {
      dup2(data->fdnull, STDOUT_FILENO);
      dup2(data->fdnull, STDERR_FILENO);
      setpgid(0, 0);
      if( isAggressive && nice(-20) < 0 )
         g_info("Can't renice player: %s", g_strerror(errno));
      
      execvp(argv[0], argv);
      g_warning(__FILE__": Error in execvp: %s", g_strerror(errno));
      return -1;
   }
   else if( data->pid < 0 )
   {
      g_warning(__FILE__": Error in fork: %s", g_strerror(errno));
      return -1;
   }
   
   g_strfreev(argv);
   return 0;
}

gint custom_suspend( CustomData *data )
{
   if( data->pid > 0 )
      kill(-1*(int)data->pid, SIGSTOP);
   else
      return -1;
   
   return 0;
}

gint custom_continue( CustomData *data )
{
   if( data->pid > 0 )
      kill(-1*(int)data->pid, SIGCONT);
   else
      return -1;
   
   return 0;
}

gint custom_init( CustomData *data )
{
   data->pid = -1;
   data->fdnull = open("/dev/null", O_WRONLY);
   if( data->fdnull < 0 ) return -1;
   
   return 0;
}

gint custom_close( CustomData *data )
{
   gint r;
   
   r = custom_stop(data);
   if( r < 0 ) return -1;
   
   r = close(data->fdnull);
   return r;
}

void custom_new( PlayerInterface *pi, gchar **command, const gchar *pattern )
{  
   CustomData *data;
    
   data = g_new0(CustomData, 1);
   data->command = command;
   data->pid = -1;
   data->gimme_next = pi->player_gimme_next;
   data->next_data = pi->user_data;
   data->stopforced = FALSE;
   
   pi->player_play = (PlayFunc)custom_play;
   pi->player_stop = (StopFunc)custom_stop;
   pi->player_suspend = (SuspendFunc)custom_suspend;
   pi->player_continue = (ContinueFunc)custom_continue;
   pi->player_init = (InitFunc)custom_init;
   pi->player_close = (CloseFunc)custom_close;
   pi->data = (gpointer)data;
   pi->pattern = g_strdup(pattern);
   
   if(loop == NULL)
   {  g_debug(__FILE__": initializing loop");
      loop = g_main_new(FALSE);
   }
}
