/* filelist.c */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
//#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
//#include <time.h>
#include <regex.h>
//#include <glib.h>

#include "config.h"
#include "logger.h"
#include "filelist.h"


char* reldir2abs( const char *reldir )
{
   static char buf[PATH_MAX];
   
   if(chdir(reldir) < 0)
   {  /* dir seems not to exist */
      return NULL;
   }
   
   if(getcwd(buf, sizeof(buf)-1) == NULL)
   {  g_critical(__FILE__": Error in getcwd: %s", g_strerror(errno));
      return NULL;
   }
   
   return buf;
}

void filelist_register_pattern( FileListData *data, player_type player, const gchar *pattern )
{
   if( pattern != NULL )
   {
      g_free(data->pattern[player]);
      data->pattern[player] = g_strdup(pattern);
   }
}

void filelist_regerror( const regex_t *preg, gint errnr )
{
   gchar *buf;
   size_t n;
   
   n = regerror(errnr, preg, NULL, 0);
   buf = g_malloc(n);
   regerror(errnr, preg, buf, n);
   
   g_info("regerror: %s", buf);
   g_free(buf);
}

int filelist_add_files( FileListData *data, player_type player, const char *pattern )
{
   DIR *d;
   struct dirent *dent;
   struct stat statbuf;
   regex_t preg;
   char *absname;
   filelist_element *fle;
   gint r;
      
   if( (r=regcomp(&preg, pattern, REG_ICASE|REG_NOSUB|REG_EXTENDED)) != 0 )
   {  filelist_regerror(&preg, r);
      return -1;
   }
   
   if( (d=opendir(data->dirname)) == NULL)
   {  g_info("Error in opening dir '%s': %s", data->dirname, g_strerror(errno));
      return -1;
   }
   
   while( (dent=readdir(d)) != NULL )
   {
      if( regexec(&preg, dent->d_name, 0, NULL, 0) == 0 )
      {
         absname = g_strconcat(data->dirname, G_DIR_SEPARATOR_S, dent->d_name, NULL);
         if( stat(absname, &statbuf) < 0)
	 {  g_warning(__FILE__": Error in stat: %s", g_strerror(errno));
	    return -1;
	 }
	 
	 if( !S_ISDIR(statbuf.st_mode) )
	 {
	    fle = g_malloc(sizeof(filelist_element));
	    fle->filename = g_strdup(dent->d_name);
	    fle->isdir = FALSE;
            fle->islist = FALSE;
            fle->player = player;
	    data->list = g_list_append(data->list, fle);
	 }
	 g_free(absname);
      }
   }
   
   if( closedir(d) < 0)
   {  g_warning(__FILE__": Error in closedir: %s", g_strerror(errno));
      return -1;
   }
   
   regfree(&preg);
   
   return 0;
}

int filelist_add_dirs( FileListData *data )
{
   DIR *d;
   struct dirent *dent;
   struct stat statbuf;
   regex_t preg;
   char *absname;
   filelist_element *fle;
   gint r;
   
   if( (r=regcomp(&preg, "^[^.].*$|^\\.\\.$", REG_NOSUB|REG_EXTENDED)) != 0 )
   {  filelist_regerror(&preg, r);
      return -1;
   }
   
   if( (d=opendir(data->dirname)) == NULL)
   {  g_info("Error in opening dir '%s': %s", data->dirname, g_strerror(errno));
      return -1;
   }
   
   while( (dent=readdir(d)) != NULL )
   {
      if( regexec(&preg, dent->d_name, 0, NULL, 0) == 0)
      {  
         absname = g_strconcat(data->dirname, G_DIR_SEPARATOR_S, dent->d_name, NULL);
         if( stat(absname, &statbuf) < 0)
	 {  g_warning(__FILE__": Error in stat: %s", g_strerror(errno));
	    return -1;
	 }
	 
	 if( S_ISDIR(statbuf.st_mode) )
	 {
	    fle = g_malloc(sizeof(filelist_element));
	    fle->filename = g_strconcat(dent->d_name, G_DIR_SEPARATOR_S, NULL);
	    fle->isdir = TRUE;
            fle->islist = FALSE;
            fle->player = PLAYER_NONE;
	    data->list = g_list_append(data->list, fle);
	 }
	 
	 g_free(absname);
      }
   }
   
   regfree(&preg);
   
   if( closedir(d) < 0)
   {  g_warning(__FILE__": Error in closedir: %s", g_strerror(errno));
      return -1;
   }
   
   return 0;
}

int filelist_add_lists( FileListData *data )
{
   DIR *d;
   struct dirent *dent;
   struct stat statbuf;
   regex_t preg;
   char *absname;
   filelist_element *fle;
   gint r;
   
   if( (r=regcomp(&preg, "^.*\\.m3u$", REG_ICASE|REG_NOSUB|REG_EXTENDED)) != 0 )
   {  filelist_regerror(&preg, r);
      return -1;
   }
   
   if( (d=opendir(data->dirname)) == NULL)
   {  g_info("Error in opening dir '%s': %s", data->dirname, g_strerror(errno));
      return -1;
   }
   
   while( (dent=readdir(d)) != NULL )
   {
      if( regexec(&preg, dent->d_name, 0, NULL, 0) == 0)
      {  
         absname = g_strconcat(data->dirname, G_DIR_SEPARATOR_S, dent->d_name, NULL);
         if( stat(absname, &statbuf) < 0)
	 {  g_warning(__FILE__": Error in stat: %s", g_strerror(errno));
	    return -1;
	 }
	 
	 if( !S_ISDIR(statbuf.st_mode) )
	 {
	    fle = g_malloc(sizeof(filelist_element));
	    fle->filename = g_strdup(dent->d_name);
	    fle->isdir = FALSE;
            fle->islist = TRUE;
            fle->player = PLAYER_NONE;
	    data->list = g_list_append(data->list, fle);
	 }
	 
	 g_free(absname);
      }
   }
   
   regfree(&preg);
   
   if( closedir(d) < 0)
   {  g_warning(__FILE__": Error in closedir: %s", g_strerror(errno));
      return -1;
   }
   
   return 0;
}

gint filelist_find_e( filelist_element *fle, gchar *string )
{
   if(strcmp(fle->filename, string) == 0)
      return 0;
   
   return -1;
}

gint filelist_get_index( FileListData *data, gchar *file )
{
   GList *link;
   
   link = g_list_find_custom(data->list, file, (GCompareFunc)filelist_find_e);
   if(link == NULL) return -1;
   
   return g_list_position(data->list, link);
}

void filelist_free_e( filelist_element *fle )
{
   g_free(fle->filename);
   g_free(fle); /* is this necessary or handled by g_list_free()? */
}

void filelist_clear( FileListData *data )
{
   g_list_foreach(data->list, (GFunc)filelist_free_e, NULL);
   g_list_free(data->list);
   data->list =  NULL;
   g_free(data->dirname);
   data->dirname = NULL;
   data->olddir = -1;
   data->lastacc = NULL;
   data->nlastacc = -1;
}

void filelist_free( FileListData *data )
{
   filelist_clear(data);
   g_free(data);
}

gint filelist_compare( filelist_element *fle1, filelist_element *fle2 )
{
   /* compare two elements according to:
       - parent dir '..' is always on top
       - dirs are above files
       - dirs and files are sorted alphabetical and case sensitive
    */
   if( strcmp(fle1->filename, "../") == 0 )
      return -1;
   else if( strcmp(fle2->filename, "../") == 0 )
      return 1;
   else if(fle1->isdir == fle2->isdir)
      return strcmp(fle1->filename, fle2->filename);
   else 
   {
      if(fle1->isdir) return -1;
      else return 1;
   }
}

void filelist_add_all( FileListData *data )
{
   gint i;
   
   filelist_add_dirs(data);
   filelist_add_lists(data);
   for(i=0; i<PLAYER_LAST; i++)
      filelist_add_files(data, i, data->pattern[i]);
   
   data->list = g_list_sort( data->list, (GCompareFunc)filelist_compare );
}

filelist_element* filelist_get_element( FileListData *data, guint n )
{
   /* seems to be the same element again? */
   if(data->nlastacc >= 0 && data->nlastacc == n );
   else if( data->nlastacc >= 0 && data->nlastacc == n-1 ) /* test if we only need the next/prev element of the cached */
      data->lastacc = g_list_next(data->lastacc);
   else if(data->nlastacc >= 0 && data->nlastacc == n+1 )
      data->lastacc = g_list_previous(data->lastacc);
   else 
      data->lastacc = g_list_nth(data->list, n);
   
   if(data->lastacc == NULL) return NULL;
   
   data->nlastacc = n;
   
   return (filelist_element*)data->lastacc->data;
}

gboolean filelist_nth_isfile( FileListData *data, guint n )
{
   filelist_element *fle;
   
   fle = filelist_get_element(data, n);
   if(fle == NULL) return FALSE;
   
   return (!fle->isdir && !fle->islist);
}

gboolean filelist_nth_isdir( FileListData *data, guint n )
{
   filelist_element *fle;
   
   fle = filelist_get_element(data, n);
   if(fle == NULL) return FALSE;
   
   return fle->isdir;
}

gboolean filelist_nth_islist( FileListData *data, guint n )
{
   filelist_element *fle;
   
   fle = filelist_get_element(data, n);
   if(fle == NULL) return FALSE;
   
   return fle->islist;
}

char* filelist_get_nth_name( FileListData *data, guint n )
{
   filelist_element *fle;
   
   fle = filelist_get_element(data, n);
   if(fle == NULL) return NULL;
   
   return fle->filename;
}

char* filelist_get_nth_absname( FileListData *data, guint n )
{
   filelist_element *fle;
   
   fle = filelist_get_element(data, n);
   if(fle == NULL) return NULL;
   
   return g_strconcat(data->dirname, G_DIR_SEPARATOR_S, fle->filename, NULL);
}

player_type filelist_get_nth_type( FileListData *data, guint n )
{
   filelist_element *fle;
   
   fle = filelist_get_element(data, n);
   if(fle == NULL) return PLAYER_NONE;
   
   return fle->player;
}

guint filelist_get_length( FileListData *data )
{
   return g_list_length(data->list);
}

char* filelist_get_dirname( FileListData *data )
{
   return data->dirname;
}

gint filelist_get_olddir( FileListData *data )
{
   return data->olddir;
}

gint filelist_open_dir( FileListData *data, const char* dirname )
{
   gchar *olddirname;
   gchar *newdirname;
   gchar *chkdirname;
   
   g_return_val_if_fail(data != NULL, -1);
   g_return_val_if_fail(dirname != NULL, -1);
   
   g_debug(__FILE__": got dir '%s'", dirname);
   
   newdirname = g_strdup(dirname);
   /* remove last '/' if appended and not root dir '/' itself */
   if( strlen(dirname) > 1 && newdirname[strlen(dirname)-1] == G_DIR_SEPARATOR)
      newdirname[strlen(dirname)-1] = '\0';
   
   if(strcmp(g_basename(newdirname), "..") == 0 )
   {  /* change into parent dir requested */
      olddirname = g_strconcat(g_basename(data->dirname), G_DIR_SEPARATOR_S, NULL);
   } else {
      olddirname = NULL;
      data->olddir = -1;
   }
      
   chkdirname = reldir2abs(newdirname);
   
   if(chkdirname == NULL) /* dir seems not to exist */
      return -1;
   
   g_debug(__FILE__": will change to dir '%s'", chkdirname);
   
   filelist_clear(data);
   
   data->dirname = g_strdup(chkdirname);
   
   filelist_add_all(data);
   
   if(olddirname != NULL)
   {  /* now we search for the index */
      data->olddir = filelist_get_index(data, olddirname);
      g_free(olddirname);
   }
   
   g_free(newdirname);
   
   return 0;
}

gint filelist_add_dir_r( FileListData *data, const gchar *dirname, PlaylistAddFunc addfunc, gpointer funcdata )
{
   gint i, n;
   gchar *odir;
   gchar *rdir;
   
   /* we save the dirname here to restore it after adding all
      the files in the dir. changing to '..' afterwards
      will not work on linked dirs! */
   odir = g_strdup(filelist_get_dirname(data));
   
   if( filelist_open_dir(data, dirname) < 0 )
   {  g_info("Error adding dir '%s' recursively: %s", dirname, g_strerror(errno));
      return -1;
   }
   
   n = filelist_get_length(data);
   
   /* parent dir has index 0 so we start from index 1 */
   for(i=1; i<n; i++)
   {
      if( filelist_nth_isfile(data, i) )
         addfunc(funcdata,
                      filelist_get_dirname(data),
                      filelist_get_nth_name(data, i),
                      filelist_get_nth_type(data, i));
      else if( filelist_nth_isdir(data, i) )
      {
         rdir = g_strdup(filelist_get_nth_name(data, i));
         filelist_add_dir_r(data, rdir, addfunc, funcdata);
         g_free(rdir);
      }
   }
   
   if( filelist_open_dir(data, odir) < 0 )
   {  g_info("Error in leaving dir: %s", g_strerror(errno));
      return -1;
   }
   g_free(odir);
   
   return 0;
}

gint filelist_open_list( FileListData *data, const gchar *filename, PlaylistAddFunc addfunc, gpointer funcdata )
{
   FILE *f;
   gchar buf[PATH_MAX+NAME_MAX+1];
   gchar *dirname, *basename;
   regex_t preg[PLAYER_LAST];
   gint i, r;
   
   for(i=0; i<PLAYER_LAST; i++)
   {
      if( (r=regcomp(&preg[i], data->pattern[i], REG_ICASE|REG_NOSUB|REG_EXTENDED)) != 0 )
      {  filelist_regerror(&preg[i], r);
         return -1;
      }
   }
   
   if( (f=fopen(filename, "r")) == NULL )
   {  g_warning("Cannot open playlist file '%s': %s", filename, g_strerror(errno));
      return -1;
   }
   
   while( fgets(buf, sizeof(buf), f) != NULL )
   {
      g_strdelimit(buf, "\n", '\0');
      
      i=0;
      while( i < PLAYER_LAST && regexec(&preg[i], buf, 0, NULL, 0) != 0 )
         i++;
      if( i == PLAYER_LAST ) continue;
      
      dirname = g_dirname(buf);
      basename = g_basename(buf);
      
      g_debug(__FILE__": found '%s/%s' (%d)", dirname, basename, i);
      if( addfunc != NULL )
         addfunc(funcdata, dirname, basename, i);
      
      g_free(dirname);
   }
   
   for(i=0; i<PLAYER_LAST; i++)
      regfree(&preg[i]);
   
   fclose(f);
   return 0;
}

gint filelist_open_unknown( FileListData *data, const gchar *filename, PlaylistAddFunc addfunc, gpointer funcdata )
{
   gchar *dirname, *basename;
   gchar *olddirname, *newdirname;
   regex_t preg[PLAYER_LAST];
   regex_t preglist;
   struct stat statbuf;
   gint i, r;
   
   /* first test if we have a directory */
   if( stat(filename, &statbuf) < 0)
   {  g_warning(__FILE__": Error in stat: %s", g_strerror(errno));
      return -1;
   }
 
   if( S_ISDIR(statbuf.st_mode) )
   {
      g_debug(__FILE__": '%s' identified as a dir", filename);
      filelist_add_dir_r(data, filename, addfunc, funcdata);
      return 0;
   }
   
   /* now try if file is a playlist */
   if( (r=regcomp(&preglist, "^.*\\.m3u$", REG_ICASE|REG_NOSUB|REG_EXTENDED)) != 0 )
   {  filelist_regerror(&preg[i], r);
      return -1;
   }

   if( regexec(&preglist, filename, 0, NULL, 0) == 0 ) /* is a list */
   {
      g_debug(__FILE__": '%s' identified as a list", filename);
      r = filelist_open_list(data, filename, addfunc, funcdata);
      regfree(&preglist);
      return r;
   }
   
   /* last chance: is it a file? and for which player? */
   for(i=0; i<PLAYER_LAST; i++)
   {
      if( (r=regcomp(&preg[i], data->pattern[i], REG_ICASE|REG_NOSUB|REG_EXTENDED)) != 0 )
      {  filelist_regerror(&preg[i], r);
         return -1;
      }
   }
   
   dirname = g_dirname(filename);
   basename = g_basename(filename);
 
   if( !g_path_is_absolute(filename) )
   {
      olddirname = g_get_current_dir();
      newdirname = g_strdup(reldir2abs(dirname));
      g_free(dirname);
      dirname = newdirname;
      chdir(olddirname);
      g_free(olddirname);
   }
   
   i=0;
   while( i < PLAYER_LAST && regexec(&preg[i], filename, 0, NULL, 0) != 0 )
      i++;

   if( i != PLAYER_LAST )
   {
      g_debug(__FILE__": identified '%s' as a file (%d)", filename, i);
      if( addfunc != NULL )
         addfunc(funcdata, dirname, basename, i);
   }
   
   g_free(dirname);

   for(i=0; i<PLAYER_LAST; i++)
      regfree(&preg[i]);
   regfree(&preglist);
   
   return 0;
}

FileListData* filelist_new()
{
   FileListData *data;
   gint i;
   
   data = g_malloc0(sizeof(FileListData));
   data->list = NULL;
   data->dirname = NULL;
   data->olddir = -1;
   for(i=0; i<PLAYER_LAST; i++)
      data->pattern[i] = g_strdup("^$");
   data->lastacc = NULL;
   data->nlastacc = -1;
   
   return data;
}

/*--------------------------------------------------------*/

#ifdef SAL
int main(int argc, char *argv[])
{
   FileListData *data;
   int i, n;
   
   if(argc != 2)
   {  printf("Usage: %s <dir>\n", argv[0]);
      exit(-1);
   }
   
   srand(time(NULL)+getpid());
   
   data = filelist_new();
   filelist_open_dir(data,argv[1]);
   
   filelist_print(data);
   printf("---\n");
   filelist_shuffle(data);
   filelist_print(data);
   
   printf("%d\n", filelist_get_index(data, ".."));
   
   filelist_free(data);
   return 0;
}
#endif /* SAL */
