/* mcplay.c */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <curses.h>
#include <signal.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
#include <glib.h>
#ifdef HAVE_LIRC
#include <lirc/lirc_client.h>
#endif

#include "config.h"
#include "logger.h"
#include "screen.h"
#include "mixer.h"
#include "player.h"

#include "mcplay.h"

/* set command options defaults */
gboolean isXterm = FALSE;
gboolean useColors = FALSE;
gchar *logfilename = NULL;
gchar *lircfilename = NULL;
gboolean isAggressive = FALSE;
gboolean startRepeat = FALSE;
gboolean startRandom = FALSE;
gboolean startPlaying = FALSE;

static ScreenData *data;
static MixerData *mixer;
static PlayerData *player;
static GMainLoop *main_loop;

#ifdef HAVE_LIRC
gboolean process_ir( GIOChannel *io, GIOCondition cond, struct lirc_config *config )
{
   char *code;
   char *cmd;
   char **c;
   int r, i;
   screen_mode mode;
   
   mode = screen_get_mode(data);
   
   if( lirc_nextcode(&code) < 0)
      return FALSE;
   
   if( code == NULL ) return TRUE;
   
   while( (r=lirc_code2char(config,code,&cmd)) == 0 && cmd != NULL )
   {
      c = g_strsplit(cmd, " ", -1);
      
      if( mode == SCREEN_FILE && strcmp(c[0], "up") == 0 )
         screen_filelist_up(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "up") == 0 )
         screen_playlist_up(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "down") == 0 )
         screen_filelist_down(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "down") == 0 )
         screen_playlist_down(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "pageup") == 0 )
         screen_filelist_pageup(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "pageup") == 0 )
         screen_playlist_pageup(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "pagedown") == 0 )
         screen_filelist_pagedown(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "pagedown") == 0 )
         screen_playlist_pagedown(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "home") == 0 )
         screen_filelist_home(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "home") == 0 )
         screen_playlist_home(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "end") == 0 )
         screen_filelist_end(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "end") == 0 )
         screen_playlist_end(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "mark") == 0 )
         screen_filelist_mark(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "mark") == 0 )
         screen_playlist_mark(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "select") == 0 )
         screen_filelist_select(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "select") == 0 )
         screen_playlist_select(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "shuffle") == 0 )
         screen_playlist_shuffle(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "sort") == 0 )
         screen_playlist_sort(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "repeat") == 0 )
         screen_playlist_toggle_repeat(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "random") == 0 )
         screen_playlist_toggle_random(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "delete") == 0 )
         screen_playlist_delete(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "move") == 0 )
         screen_playlist_move(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "markall") == 0 )
         screen_playlist_mark_all(data);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "clearall") == 0 )
         screen_playlist_unmark_all(data);
      else if( mode == SCREEN_FILE && strcmp(c[0], "toggle") == 0 )
         screen_set_mode(data, SCREEN_PLAY);
      else if( mode == SCREEN_PLAY && strcmp(c[0], "toggle") == 0 )
         screen_set_mode(data, SCREEN_FILE);
      else if( strcmp(c[0], "next") == 0 )
         screen_player_next(data);
      else if( strcmp(c[0], "prev") == 0 )
         screen_player_prev(data);
      else if( strcmp(c[0], "pause") == 0 )
         screen_player_toggle_pause(data);
      else if( strcmp(c[0], "stop") == 0 )
         screen_player_stop(data);
      else if( strcmp(c[0], "counter") == 0 )
         screen_toggle_timer(data);
      else if( strcmp(c[0], "seek") == 0 )
         screen_player_seek(data, g_strtod(c[1], NULL));
      else if( strcmp(c[0], "vol") == 0 )
      {
         g_return_val_if_fail(c[1]!=NULL, TRUE);
         i = (gint)g_strtod(c[1], NULL);
         if( c[1][0] == '+' || c[1][0] == '-' )
            i=mixer_inc_volume(mixer, i);
         else
            mixer_set_volume(mixer, i);
         screen_message(data, 2, "Volume %d%%", i);
      }
      else if( strcmp(c[0], "quit") == 0 )
         g_main_quit(main_loop);
      else
         g_warning("Unknown ir command (or in wrong mode): '%s'", c[0]);
      
      g_strfreev(c);
   }
   
   g_free(code);
   
   if(r==-1)
   {  g_warning("Error processing lirc config file");
      return FALSE;
   }
   
   return TRUE;
}
#endif /* HAVE_LIRC */

gboolean process_key( GIOChannel *io, GIOCondition cond, ScreenData *data )
{
   gboolean global = FALSE;
   static screen_mode oldmode = SCREEN_FILE;
   static gint oldkey = -1;
   gchar *answer = NULL;
   gint key, r;
      
   if( cond != G_IO_IN )
   {  g_error("Error in reading from standard input");
      return FALSE;
   }
      
   if( oldkey >= 0 )
   {
      answer = screen_ask_getch(data);
      if( answer == NULL )
         return TRUE;
      else if( strlen(answer) < 1 )
      {  oldkey = -1;
         return TRUE;
      }
      
      g_debug(__FILE__": got answer '%s'", answer);
      
      key = oldkey;
      oldkey = -1;
   }
   else
      key = getch();
      
   if(data->mode == SCREEN_FILE)
   {  switch(key)
      {
         case KEY_UP:
         case 'k':
            screen_filelist_up(data);
            break;
         case KEY_DOWN:
         case 'j':
            screen_filelist_down(data);
            break;
         case KEY_PPAGE:
         case 'K':
            screen_filelist_pageup(data);
            break;
         case KEY_NPAGE:
         case 'J':
            screen_filelist_pagedown(data);
            break;
         case KEY_HOME:
         case 'g':
            screen_filelist_home(data);
            break;
         case KEY_END:
         case 'G':
            screen_filelist_end(data);
            break;
         case KEY_BACKSPACE:
            screen_filelist_chng_dir(data, "../");
            break;
         case 'o':
            if(answer)
               screen_filelist_chng_dir(data, answer);
            else { oldkey = key; screen_ask_start(data, "open: "); }
            break;
         case '\t':
            screen_set_mode(data, SCREEN_PLAY);
            break;
         case '\r':
         case KEY_ENTER:
            screen_filelist_select(data);
            break;
         case ' ':
            screen_filelist_mark(data);
            /* screen_filelist_down(data); */
            break;
         default:
            global = TRUE;
      }
   }
   else if(data->mode == SCREEN_PLAY)
   {  switch(key)
      {
         case KEY_UP:
         case 'k':
            screen_playlist_up(data);
            break;
         case KEY_DOWN:
         case 'j':
            screen_playlist_down(data);
            break;
         case KEY_PPAGE:
         case 'K':
            screen_playlist_pageup(data);
            break;
         case KEY_NPAGE:
         case 'J':
            screen_playlist_pagedown(data);
            break;
         case KEY_HOME:
         case 'g':
            screen_playlist_home(data);
            break;
         case KEY_END:
         case 'G':
            screen_playlist_end(data);
            break;
         case '\r':
         case KEY_ENTER:
            screen_playlist_select(data);
            break;
         case ' ':
            screen_playlist_mark(data);
            /* screen_playlist_down(data); */
            break;
         case 'd':
         case KEY_DC:
            screen_playlist_delete(data);
            break;
         case 'm':
            screen_playlist_move(data);
            break;
         case 's':
            screen_playlist_shuffle(data);
            break;
         case 'S':
            screen_playlist_sort(data);
            break;
         case 'r':
            screen_playlist_toggle_repeat(data);
            break;
         case 'R':
            screen_playlist_toggle_random(data);
            break;
         case 'a':
            screen_playlist_mark_all(data);
            break;
         case 'c':
            screen_playlist_unmark_all(data);
            break;
         case 'A':
            if(answer)
               screen_playlist_mark_reg(data, answer);
            else { oldkey = key; screen_ask_start(data, "mark regex: "); }
            break;
         case 'C':
            if(answer)
               screen_playlist_unmark_reg(data, answer);
            else { oldkey = key; screen_ask_start(data, "clear regex: "); }
            break;
         case 'w':
            if(answer)
               screen_playlist_write(data, answer);
            else { oldkey = key; screen_ask_start(data, "write playlist: "); }
            break;
         case '\t':
            screen_set_mode(data, SCREEN_FILE);
            break;
         default:
            global = TRUE;
      }
   }
   else if(data->mode == SCREEN_HELP)
   {
      screen_set_mode(data, oldmode); /* restore old mode */
   }
   
   if(global) /* global keys */
   {  switch(key)
      {
         case 'q':
            g_main_quit(main_loop);
            break;
         case 'h':
            oldmode = screen_get_mode(data); /* save old mode to restore later */
            screen_set_mode(data, SCREEN_HELP);
            break;
         case 'x':
            screen_player_stop(data);
            break;
         case 'z':
            screen_player_toggle_pause(data);
            break;
         case 'n':
            screen_player_next(data);
            break;
         case 'p':
            screen_player_prev(data);
            break;
         case KEY_LEFT:
         case '\002':
            screen_player_seek(data, -1.5);
            break;
         case KEY_RIGHT:
         case '\006':
            screen_player_seek(data, +1.5);
            break;
         case 't':
            screen_toggle_timer(data);
            break;
         case '0':
            mixer_set_volume(mixer, 0);
            screen_message(data, 2, "Volume %d%%", 0);
            break;
         case '1':
            mixer_set_volume(mixer, 10);
            screen_message(data, 2, "Volume %d%%", 10);
            break;
         case '2':
            mixer_set_volume(mixer, 20);
            screen_message(data, 2, "Volume %d%%", 20);
            break;
         case '3':
            mixer_set_volume(mixer, 30);
            screen_message(data, 2, "Volume %d%%", 30);
            break;
         case '4':
            mixer_set_volume(mixer, 40);
            screen_message(data, 2, "Volume %d%%", 40);
            break;
         case '5':
            mixer_set_volume(mixer, 50);
            screen_message(data, 2, "Volume %d%%", 50);
            break;
         case '6':
            mixer_set_volume(mixer, 60);
            screen_message(data, 2, "Volume %d%%", 60);
            break;
         case '7':
            mixer_set_volume(mixer, 70);
            screen_message(data, 2, "Volume %d%%", 70);
            break;
         case '8':
            mixer_set_volume(mixer, 80);
            screen_message(data, 2, "Volume %d%%", 80);
            break;
         case '9':
            mixer_set_volume(mixer, 90);
            screen_message(data, 2, "Volume %d%%", 90);
            break;
         case '+':
            r=mixer_inc_volume(mixer, 2);
            screen_message(data, 2, "Volume %d%%", r);
            break;
         case '-':
            r=mixer_inc_volume(mixer, -2);
            screen_message(data, 2, "Volume %d%%", r);
            break;
         case '\014':
            screen_hardrefresh(data);
            break;
         case KEY_RESIZE:
            g_debug(__FILE__": got KEY_RESIZE");
            /* screen_resize(data); */
            break;
      }
   }
   
   return TRUE;
}

gboolean mcplay_resize_timeout( ScreenData *data )
{
   g_debug(__FILE__": resize_timeout called");
   screen_resize(data);
   
   return FALSE;
}

void mcplay_sigresize( int sig )
{
/*    struct winsize win;
   
   ioctl(STDIN_FILENO, TIOCGWINSZ, &win);
   
   if (win.ws_col > 10 && win.ws_row > 5)
   {
      resizeterm(win.ws_row, win.ws_col);
      g_debug(__FILE__": resized to %d, %d", win.ws_row, win.ws_col);
   }
   
   refresh();
   signal(SIGWINCH, mcplay_sigresize); */
   
   g_timeout_add( 1, (GSourceFunc)mcplay_resize_timeout, data);
}

int mcplay_run( gchar *optfiles[] )
{
   GIOChannel *io;
#ifdef HAVE_LIRC
   gint lircfd;
   struct lirc_config *lconf;
#endif
   gchar *s, **o;
   time_t tim;
   
   time(&tim);
   srand(tim+getpid());
   
   logger_open(logfilename);
   g_message("----------------------------------------\n"
             "restarted "PROGNAME_S" "VERSION_S" on %s", ctime(&tim));

   if( (s=getenv("TERM")) && strncmp(s, "xterm", 5)==0 )
      isXterm = TRUE;
   
   if(isXterm)
   {  fprintf(stderr, "\033]0;mcplay\a");
      fflush(stderr);
   }
   
   data = screen_new();
   mixer = mixer_new();
   player = player_new();
   screen_set_player(data, player);
   
   signal(SIGWINCH, mcplay_sigresize);
   
   o=optfiles;
   while( *o != NULL )
   {
      g_debug(__FILE__": got optfile '%s'", *o);
      screen_playlist_add_unknown(data, *o);
      o++;
   }
   
   screen_message(data, 2, "Welcome to "PROGNAME_S" "VERSION_S);
   
   if(startRepeat)
      screen_playlist_toggle_repeat(data);
   if(startRandom)
      screen_playlist_toggle_random(data);
   if(startPlaying)
   {  screen_set_mode(data, SCREEN_PLAY);
      screen_playlist_select(data);
   }
   
   main_loop = g_main_new(TRUE);
   
   io = g_io_channel_unix_new(STDIN_FILENO);
   g_io_add_watch(io, G_IO_IN|G_IO_ERR|G_IO_HUP, (GIOFunc)process_key, data);

#ifdef HAVE_LIRC
   lircfd = lirc_init(PROGNAME_S, 1);
   if(lircfd < 0)
      g_critical("Error in initialising lirc");
   else
   {  
      lirc_readconfig(lircfilename, &lconf, NULL);
      io = g_io_channel_unix_new(lircfd);
      g_io_add_watch(io, G_IO_IN|G_IO_ERR|G_IO_HUP, (GIOFunc)process_ir, lconf);
   }
#endif
   
   g_main_run(main_loop);

#ifdef HAVE_LIRC
   lirc_freeconfig(lconf);
   lirc_deinit();
#endif

   player_free(player);
   mixer_free(mixer);

   if(isXterm)
   {  fprintf(stderr, "\033]0;xterm\a");
      fflush(stderr);
   }
   
   logger_close();
   
   return 0;
}

void print_help(void)
{
   g_print(PROGNAME_S", version "VERSION_S"\n"
           "Copyright (C) 2002 Tobias Rapp (yahuxo@gmx.de).\n"
           "\n"
           "Usage: "PROGNAME_S" [options] [file|dir|playlist]...\n"
           "\n"
           "Files, Directories and Playlist-Files (m3u-files) given as arguments on the\n"
           "commandline will be added to the internal playlist.\n"
           "\n"
           "Options are (defaults in brackets):\n"
           "  -a, --aggressive  Increase the priority of the players (if possible).\n"
           "  -c, --color       Run program in colored mode (if available).\n"
           "  -l, --logfile     Log messages to this file.  ["LOGFILE_S"]\n"
#ifdef HAVE_LIRC
           "  -L, --lircrc      Load this lirc config file.  [~/.lircrc]\n"
#endif
           "  -p, --play        Start playing the playlist.\n"
           "  -r, --repeat      Start in repeat playlist mode.\n"
           "  -R, --random      Start in random playlist mode.\n"
           "      --version     Print version information.\n"
           "      --copyright   Print copyright information.\n"
           "  -h, --help        Shows this text.\n"
           "\n"
           "Press 'h' while "PROGNAME_S" is running for help on keys.\n");
}

void print_version(void)
{
   g_print("This is "PROGNAME_S", version "VERSION_S" (compiled %s).\n", __DATE__);
   g_print("Compiled in support for:\n"
#ifdef HAVE_MPG123
           "  mpg123\n"
#endif
#ifdef HAVE_OGG123
           "  ogg123\n"
#endif
#ifdef HAVE_SOX
           "  sox\n"
#endif
#ifdef HAVE_DRVMIDI
           "  drvmidi\n"
#endif
#ifdef HAVE_PLAYMIDI
           "  playmidi\n"
#endif
#ifdef HAVE_LIRC
           "  lirc\n"
#endif
           );
}

void print_copyright(void)
{
   g_print(PROGNAME_S", version "VERSION_S"\n"
           "Copyright (C) 2002 Tobias Rapp.\n"
           "\n"
           "This program is free software; you can redistribute it and/or modify\n"
           "it under the terms of the GNU General Public License as published by\n"
           "the Free Software Foundation; either version 2 of the License, or\n"
           "(at your option) any later version.\n"
           "\n"
           "This program is distributed in the hope that it will be useful,\n"
           "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
           "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
           "GNU General Public License for more details.\n"
           "\n"
           "You should have received a copy of the GNU General Public License\n"
           "along with this program. If not, write to the Free Software\n"
           "Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n"
           "\n");
}

int main(int argc, char *argv[])
{
   struct option long_options[] =
   {
      { "aggressive", 0, NULL, 'a' },
      { "color",      0, NULL, 'c' },
      { "copyright",  0, NULL, 'C' },
      { "logfile",    1, NULL, 'l' },
      { "lircrc",     1, NULL, 'L' },
      { "play",       0, NULL, 'p' },
      { "repeat",     0, NULL, 'r' },
      { "random",     0, NULL, 'R' },
      { "version",    0, NULL, 'V' },
      { "help",       0, NULL, 'h' },
      { 0, 0, 0, 0 }
   };
   gchar c;
   
   logfilename = g_strdup(LOGFILE_S);
   
   while( (c=getopt_long(argc, argv, "acl:L:rRph", long_options, NULL)) != EOF )
   {
      switch(c)
      {
         case 'a':
            isAggressive = TRUE;
            break;
         case 'c':
            useColors = TRUE;
            break;
         case 'l':
            g_free(logfilename);
            logfilename = g_strdup(optarg);
            break;
         case 'L':
            lircfilename = g_strdup(optarg);
            break;
         case 'p':
            startPlaying = TRUE;
            break;
         case 'r':
            startRepeat = TRUE;
            break;
         case 'R':
            startRandom = TRUE;
            break;
         case 'h':
            print_help();
            exit(0);
            break;
         case 'V':
            print_version();
            exit(0);
            break;
         case 'C':
            print_copyright();
            exit(0);
            break;
         default:
            g_print("Unknown charakter '%c'\n", c);
         case '?':
            g_print("Try '"PROGNAME_S" --help' for more information.\n");
            exit(-1);
      }
   }
   
   return mcplay_run(argv+optind);
}
