#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <glib.h>

#include "config.h"
#include "logger.h"
#include "mcplay.h"
#include "mpg123.h"


typedef struct _Mpg123Data
{
   gint pid;
   gint fdwr;
   gint fdrd;
   GIOChannel *io;
   gdouble sec_elp;
   gdouble sec_rem;
   gint frame_elp;
   gint frame_rem;
   GimmeNextFunc gimme_next;
   gpointer next_data;
} Mpg123Data;

/* private data/functions */
static Mpg123Data *sigdata;
static GMainLoop *loop;
gint mpg123_init( Mpg123Data *data );

void mpg123_sigchld( int signal )
{
   int status, r;
   
   g_debug(__FILE__": got SIGCHLD ... ");
   r=waitpid(-1, &status, WNOHANG);
   
   if( r == 0 )
   {  g_debug("just paused?");
      return;
   }
   else if( r != sigdata->pid )
   {
      g_debug("pid %d, not this player's child?", r);
      return;
   }
   
   sigdata->pid = -1;
   sigdata->sec_elp = -1;
   sigdata->sec_rem = -1;
   sigdata->frame_elp = -1;
   sigdata->frame_rem = -1;
   
   if( g_main_is_running(loop) )
      g_main_quit(loop);
   
   g_debug("ok. (pid %d, stat %d)", r, WEXITSTATUS(status));
}

gint mpg123_play( Mpg123Data *data, const char *file )
{
   gchar buf[3+PATH_MAX+NAME_MAX];
   gint r;
   
   if( data->pid < 0 )
   {  g_warning(__FILE__": player seems to have died");
      g_debug(__FILE__": will restart player");
      mpg123_init(data);
      /* return -1; */
   }
   
   data->sec_elp = 0;
   data->sec_rem = 0;
   
   g_snprintf(buf, sizeof(buf)-1, "L %s\n", file);
   r = write(data->fdwr, buf, strlen(buf));
   
   return r-strlen(buf);
}

gint mpg123_stop( Mpg123Data *data )
{   
   gchar buf[8];
   gint r;
   
   if( data->pid < 0 )
      return -1;

   if( data->sec_elp < 0 || data->sec_rem < 0 )
      return -1;

   g_snprintf(buf, sizeof(buf)-1, "S\n");
   r = write(data->fdwr, buf, strlen(buf));
   
   return r-strlen(buf);
}

gint mpg123_pause( Mpg123Data *data )
{
   gchar buf[8];
   gint r;
   
   if( data->pid < 0 )
      return -1;
   
   if( data->sec_elp < 0 || data->sec_rem < 0 )
      return -1;
   
   g_snprintf(buf, sizeof(buf)-1, "P\n");
   r = write(data->fdwr, buf, strlen(buf));
   
   return r-strlen(buf);
}

gint mpg123_seek( Mpg123Data *data, gdouble sec )
{
   gchar buf[32];
   gint r;
   gdouble fps;
   
   if( data->pid < 0 )
      return -1;
   
   if( data->sec_elp < 0 || data->sec_rem < 0 )
      return -1;
   
   fps = (data->frame_elp+data->frame_rem)/(data->sec_elp+data->sec_rem);
   g_debug(__FILE__": fps %g, seeking %+d frames", fps, (gint)(sec*fps));
   
   g_snprintf(buf, sizeof(buf)-1, "J %+d\n", (gint)(sec*fps));
   
   r = write(data->fdwr, buf, strlen(buf));
   
   return r-strlen(buf);
}

gint mpg123_progress( Mpg123Data *data, gdouble *sec_elp, gdouble *sec_rem )
{
   if( data->pid < 0 )
      return -1;
   
   if( data->sec_elp < 0 || data->sec_rem < 0 )
      return -1;
   
   *sec_elp = data->sec_elp;
   *sec_rem = data->sec_rem;
   
   return 0;
}

gboolean mpg123_process_io( GIOChannel *io, GIOCondition cond, Mpg123Data *data )
{
   gchar buf[INPUT_MAX];
   gchar *s;
   guint r;
   static gboolean isWaiting = FALSE;
   static guint i = 0;
   
   if( cond != G_IO_IN )
   {
      g_debug(__FILE__": closing pipe.");
      g_io_channel_close(io);
      return FALSE;
   }
   
   if( g_io_channel_read(io, buf+i, sizeof(buf)-i-1, &r) != G_IO_ERROR_NONE )
   {
      g_warning(__FILE__": Error in read: %s", g_strerror(errno));
      return TRUE;
   }
   i += r;
   buf[i] = '\0';
   
   while( (r=strcspn(buf, "\n")) < strlen(buf) ) /* newline found */
   {
      s = g_strndup(buf, r);
      memmove(buf, buf+r+1, strlen(buf)-r); /* move stuff behind newline to front */
 
      /* g_strstrip(s); */
 
      if( strncmp(s, "@F", 2) == 0 )
      {
         sscanf(s+2, "%d %d %lf %lf", &data->frame_elp, &data->frame_rem, &data->sec_elp, &data->sec_rem);
         /* g_debug(__FILE__": @F %d %d %g %g", data->frame_elp, data->frame_rem, data->sec_elp, data->sec_rem); */
 
         if( data->frame_rem < 5 && sigdata->gimme_next != NULL && !isWaiting )
         {
            isWaiting = TRUE;
            data->sec_elp = -1;
            data->sec_rem = -1;
 
            g_debug(__FILE__": adding timeout to get new song");
            g_timeout_add( 1, sigdata->gimme_next, sigdata->next_data);
         }
         else if( data->frame_rem > 5 && isWaiting )
         {
	    g_debug(__FILE__": cleared waiting-for-next-song flag");
            isWaiting = FALSE;
         }
      }
      else if( strncmp(s, "@P 0", 4) == 0 )
      {
         data->sec_elp = -1;
         data->sec_rem = -1;
      }
      else
         g_debug(__FILE__": >%s<", s);
      
      g_free(s);
   }
   
   i = strlen(buf);
   
   return TRUE;
}

gint mpg123_init( Mpg123Data *data )
{
   /* GIOChannel *io; */
   gint fdwr[2], fdrd[2];
   
   if( data->pid > 0 )
      return 0;
   
   if( pipe(fdwr) < 0 || pipe(fdrd) < 0 )
   {  g_warning(__FILE__": Error in creating pipes: %s", g_strerror(errno));
      return -1;
   }
   
   signal(SIGPIPE, SIG_IGN);
   sigdata = data;
   signal(SIGCHLD, mpg123_sigchld);
   
   if( (data->pid=fork()) == 0 )
   {
      dup2(fdwr[0], STDIN_FILENO);
      close(fdwr[0]);
      close(fdwr[1]);
      dup2(fdrd[1], STDOUT_FILENO);
      /* dup2(fdrd[1], STDERR_FILENO); */
      close(STDERR_FILENO);
      close(fdrd[0]);
      close(fdrd[1]);
      
      if( isAggressive && nice(-20) < 0 )
         g_info("Can't renice player: %s", g_strerror(errno));
      
      execlp("mpg123", "mpg123", "-R", /* "--resync", */ /* "-b", "500", */ "-a", DSP_DEV, /* "--aggressive", */ "-", NULL);
      
      /* never reached if exec successful */
      g_warning(__FILE__": Error in execlp: %s", g_strerror(errno));
      return -1;
   }
   else if( data->pid < 0 )
   {
      g_warning(__FILE__": Error in fork: %s", g_strerror(errno));
      return -1;
   }
   
   close(fdwr[0]);
   data->fdwr = fdwr[1];
   data->fdrd = fdrd[0];
   close(fdrd[1]);
   
   data->io = g_io_channel_unix_new(data->fdrd);
   g_io_add_watch(data->io, G_IO_IN|G_IO_ERR|G_IO_HUP, (GIOFunc)mpg123_process_io, data);
   
   return 0;
}

gint mpg123_close( Mpg123Data *data )
{
   gchar buf[8];
   gint r;
   
   if( data->pid < 0 )
      return -1;
   
   g_snprintf(buf, sizeof(buf)-1, "Q\n");
   r = write(data->fdwr, buf, strlen(buf));
   
   g_debug(__FILE__": mpg123_close()");
   
   g_main_run(loop); /* we wait here until the signal arrives that our child is dead :-( */
   
   g_io_channel_close(data->io);
   
   close(data->fdwr);
   data->fdwr = -1;
   close(data->fdrd);
   data->fdrd = -1;
   
   return 0;
}

void mpg123_new( PlayerInterface *pi )
{  
   Mpg123Data *data;
    
   data = g_new0(Mpg123Data, 1);
   data->pid = -1;
   data->fdwr = -1;
   data->fdrd = -1;
   data->sec_elp = -1;
   data->sec_rem = -1;
   data->frame_elp = -1;
   data->frame_rem = -1;
   data->gimme_next = pi->player_gimme_next;
   data->next_data = pi->user_data;
   
   pi->player_play = (PlayFunc)mpg123_play;
   pi->player_stop = (StopFunc)mpg123_stop;
   pi->player_suspend = (SuspendFunc)mpg123_pause;
   pi->player_continue = (ContinueFunc)mpg123_pause;
   pi->player_seek = (SeekFunc)mpg123_seek;
   pi->player_progress = (ProgressFunc)mpg123_progress;
   pi->player_init = (InitFunc)mpg123_init;
   pi->player_close = (CloseFunc)mpg123_close;
   pi->data = (gpointer)data;
   pi->pattern = "^.*\\.mp[123]$";
   
   if( loop == NULL )
   {  g_debug(__FILE__": initialising loop");
      loop = g_main_new(FALSE);
   }
}
