#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <glib.h>

#include "config.h"
#include "logger.h"
#include "mcplay.h"
#include "ogg123.h"

typedef struct _Ogg123Data
{
   gint pid;
   gint fdrd;
   
   GimmeNextFunc gimme_next;
   gpointer next_data;
   
   gboolean stopforced;
   gdouble sec_elp;
   gdouble sec_rem;
} Ogg123Data;

static Ogg123Data *sigdata = NULL;
static GMainLoop *loop = NULL;

void ogg123_sigchld( int signal )
{
   int status, r;
   
   g_debug(__FILE__": got SIGCHLD ... ");
   r=waitpid(-1, &status, WNOHANG);
   
   if( r == 0 )
   {  g_debug("just paused?");
      return;
   }
   else if( r != sigdata->pid )
   {  g_debug("pid %d, not this player's child?", r);
      return;
   }
   
   sigdata->pid = -1;
   sigdata->sec_elp = -1;
   sigdata->sec_rem = -1;
   
   if( sigdata->gimme_next != NULL && !sigdata->stopforced )
      g_timeout_add( 1, sigdata->gimme_next, sigdata->next_data);
   
   sigdata->stopforced = FALSE;
      
   if( g_main_is_running(loop) )
      g_main_quit(loop);
   
   g_debug("ok. (pid %d, stat %d)", r, WEXITSTATUS(status));
}

gboolean ogg123_kill_handler( Ogg123Data *data )
{
   gint r;
   
   kill(-1*data->pid, SIGCONT);
   r=kill(-1*data->pid, SIGKILL);
   
   if( r < 0 )
      g_warning(__FILE__": cannot kill process %d: %s", data->pid, g_strerror(errno));
   
   return FALSE;
}

gint ogg123_stop( Ogg123Data *data )
{
   g_debug(__FILE__": ogg123_stop()");
   
   if( data->pid > 0 )
   {
      data->stopforced = TRUE;
      
      /* we wait until the signal arrives that our child is dead :-( */
      g_timeout_add( 1, (GSourceFunc)ogg123_kill_handler, data);
      g_main_run(loop);
      g_debug(__FILE__": player stopped.");
   }
   
   return 0;
}

gint ogg123_progress( Ogg123Data *data, gdouble *sec_elp, gdouble *sec_rem )
{
   if( data->pid < 0 )
      return -1;
   
   if( data->sec_elp < 0 || data->sec_rem < 0 )
      return -1;
   
   *sec_elp = data->sec_elp;
   *sec_rem = data->sec_rem;
   
   return 0;
}

gboolean ogg123_process_io( GIOChannel *io, GIOCondition cond, Ogg123Data *data )
{
   gchar buf[INPUT_MAX];
   gchar *v, *s;
   guint r;
   gint min_elp, min_rem;
   static guint i = 0;
   static gboolean isWaiting = FALSE;
   
   if( cond != G_IO_IN )
   {
      g_debug(__FILE__": closing pipe.");
      g_io_channel_close(io);
      return FALSE;
   }
   
   if( g_io_channel_read(io, buf+i, sizeof(buf)-i-1, &r) != G_IO_ERROR_NONE )
   {
      g_warning(__FILE__": Error in read: %s", g_strerror(errno));
      return TRUE;
   }
   i += r;
   buf[i] = '\0';
   
   while( (r=strcspn(buf, "\r\n")) < strlen(buf) ) /* newline found */
   {
      s = g_strndup(buf, r);
      memmove(buf, buf+r+1, strlen(buf)-r); /* move stuff behind newline to front */
 
      g_strstrip(s);
      if( strlen(s) == 0 )
      {  g_free(s);
         continue;
      }
      
      v = strchr(s, ':') + 1;
      
      if( strncmp(s, "Time:", 5) == 0 )
      {
         sscanf(v, "%d:%lf [%d:%lf]", &min_elp, &data->sec_elp, &min_rem, &data->sec_rem);
         data->sec_elp += min_elp * 60.0;
         data->sec_rem += min_rem * 60.0;
         /* g_printerr(__FILE__": # %g %g\n", data->sec_elp, data->sec_rem); */
         
         if( data->sec_rem < 0.2 && sigdata->gimme_next != NULL && !isWaiting )
         {
            isWaiting = TRUE;
            data->sec_elp = -1;
            data->sec_rem = -1;
            
            g_debug(__FILE__": adding timeout to get new song");
            g_timeout_add( 1, sigdata->gimme_next, sigdata->next_data);
         }
         else if( data->sec_rem > 0.2 && isWaiting )
         {
            isWaiting = FALSE;
         }
      }
      else
         g_debug(__FILE__": >%s<", s);
      
      g_free(s);
   }
   
   i = strlen(buf);
   
   return TRUE;
}

gint ogg123_play( Ogg123Data *data, const gchar *file )
{
   GIOChannel *io;
   gint fdrd[2];
   
   if( data->pid > 0 )
   {  data->stopforced = TRUE;
      ogg123_stop(data);
      data->stopforced = FALSE;
   }
   
   if( pipe(fdrd) < 0 )
   {  g_warning(__FILE__": Error in creating pipe: %s", g_strerror(errno));
      return -1;
   }
   
   signal(SIGPIPE, SIG_IGN);
   sigdata = data;   
   signal(SIGCHLD, ogg123_sigchld);
   
   if( (data->pid=fork()) == 0 )
   {
      dup2(fdrd[1], STDOUT_FILENO);
      dup2(fdrd[1], STDERR_FILENO);
      close(fdrd[0]);
      close(fdrd[1]);
      
      setpgid(0, 0);
      if( isAggressive && nice(-20) < 0 )
         g_info("Can't renice player: %s", g_strerror(errno));
      
      execlp("ogg123", "ogg123", "-v", "-b", "500", "-d", "oss", "-o", "dsp:"DSP_DEV, file, NULL);
      
      /* never reached if exec successful */
      g_warning(__FILE__": Error in execlp: %s", g_strerror(errno));
      return -1;
   }
   else if( data->pid < 0 )
   {
      g_warning(__FILE__": Error in fork: %s", g_strerror(errno));
      return -1;
   }
   
   data->fdrd = fdrd[0];
   close(fdrd[1]);
   
   data->sec_elp = 0;
   data->sec_rem = 0;
   
   io = g_io_channel_unix_new(data->fdrd);
   g_io_add_watch(io, G_IO_IN|G_IO_ERR|G_IO_HUP, (GIOFunc)ogg123_process_io, data);
   
   return 0;
}

gint ogg123_suspend( Ogg123Data *data )
{
   /* g_debug(__FILE__": called player suspend"); */
   
   if( data->pid > 0 )
      kill(-1*data->pid, SIGSTOP);
   else
      return -1;
   
   return 0;
}

gint ogg123_continue( Ogg123Data *data )
{
   /* g_debug(__FILE__": called player continue"); */
   
   if( data->pid > 0 )
      kill(-1*data->pid, SIGCONT);
   else
      return -1;
   
   return 0;
}

void ogg123_new( PlayerInterface *pi )
{  
   Ogg123Data *data;
    
   data = g_new0(Ogg123Data, 1);
   data->pid = -1;
   data->fdrd = -1;
   data->gimme_next = pi->player_gimme_next;
   data->next_data = pi->user_data;
   data->stopforced = FALSE;
   
   pi->player_play = (PlayFunc)ogg123_play;
   pi->player_stop = (StopFunc)ogg123_stop;
   pi->player_suspend = (SuspendFunc)ogg123_suspend;
   pi->player_continue = (ContinueFunc)ogg123_continue;
   pi->player_progress = (ProgressFunc)ogg123_progress;
   pi->player_close = (CloseFunc)ogg123_stop;
   pi->data = (gpointer)data;
   pi->pattern = "^.*\\.ogg$";
   
   if(loop == NULL)
   {  g_debug(__FILE__": initializing loop");
      loop = g_main_new(FALSE);
   }
}

/* EOF */
