/* player.c */

#include "config.h"
#include "logger.h"
#include "player.h"

#include "custom.h"
#ifdef HAVE_MPG123
#include "mpg123.h"
#endif
#ifdef HAVE_OGG123
#include "ogg123.h"
#endif

/* private functions */
gboolean player_gimme_next( PlayerData *data );

/* custom player commands (needed by custom_new(); see below) */
/* "%s" will be substituted with the filename */
#if 0 /* HAVE_MPG123 */
static gchar *mpg123[] = { "mpg123", "-q", "-b", "500", "-a", DSP_DEV, "--aggressive", "%s", NULL };
#endif
#if 0 /* HAVE_OGG123 */
static gchar *ogg123[] = { "ogg123", "-q", "-b", "500", "-d", "oss", "-o", "dsp:"DSP_DEV, "%s", NULL };
#endif
#ifdef HAVE_SOX
static gchar *sox[] = { "sox", "%s", "-t", "ossdsp", DSP_DEV, NULL };
#endif
#ifdef HAVE_DRVMIDI
static gchar *drvmidi[] = { "drvmidi", "-i", "d", "%s", NULL };
#endif
#ifdef HAVE_PLAYMIDI
static gchar *playmidi[] = { "playmidi", "-e", "%s", NULL };
#endif


PlayerData* player_new()
{
   PlayerData *data;
   gint i;
   
   data = g_new0(PlayerData, 1);
   
   for(i=0; i<PLAYER_LAST; i++)
   {
      data->player[i].player_init = NULL;
      data->player[i].player_play = NULL;
      data->player[i].player_stop = NULL;
      data->player[i].player_suspend = NULL;
      data->player[i].player_continue = NULL;
      data->player[i].player_seek = NULL;
      data->player[i].player_progress = NULL;
      data->player[i].player_close = NULL;
      data->player[i].data = NULL;
      data->player[i].player_gimme_next = (GimmeNextFunc)player_gimme_next;
      data->player[i].user_data = data;
      data->player[i].pattern = NULL;
   }
   data->currtype = PLAYER_NONE;
   data->get_next = NULL;
   data->next_data = NULL;
   data->ispaused = FALSE;
   
#if 0
   custom_new( &data->player[PLAYER_MP3],
               mpg123,			/* "mpg123 -q -b 500 --aggressive %s" */
                "^.*\\.mp[123]$" );
#endif
#ifdef HAVE_MPG123
   mpg123_new( &data->player[PLAYER_MP3] );
#endif
#if 0
   custom_new( &data->player[PLAYER_OGG],
               ogg123, 			/* "ogg123 -q -b 500 -d oss \"%s\"" */
               "^.*\\.ogg$" );
#endif
#ifdef HAVE_OGG123
   ogg123_new( &data->player[PLAYER_OGG] );
#endif
#ifdef HAVE_SOX
   custom_new( &data->player[PLAYER_SOX],
               sox, 			/* "sox \"%s\" -t ossdsp /dev/dsp" */
               "^.*\\.(aiff|au|cdr|wav)$" );
#endif
#ifdef HAVE_DRVMIDI
   custom_new( &data->player[PLAYER_MID],
               drvmidi,
               "^.*\\.(mid|midi)$" );
#endif
#ifdef HAVE_PLAYMIDI
   cusom_new( &data->player[PLAYER_MID],
              playmidi,
              "^.*\\.(mid|midi)$" );
#endif
   
   return data;
}

void player_free( PlayerData *data )
{
   if( data->currtype != PLAYER_NONE )
   {  if( player_close(data) < 0 )
      {  g_warning(__FILE__": Error in closing player");
      }
   }
   
   g_free(data);
}

gboolean player_gimme_next( PlayerData *data )
{   
   g_debug(__FILE__": player_gimme_next() called");
   
   if( data->get_next == NULL )
      return FALSE;
   
   data->get_next( data->next_data );
      
   return FALSE;
}

gint player_init( PlayerData *data, player_type type )
{
   PlayerInterface *func;
   gint r;
   
   if( data->currtype == type )
      return 0;
   
   if( data->currtype != PLAYER_NONE )
   {  if( player_close(data) < 0 )
      {  g_warning(__FILE__": Error in closing player.");
         /* return -1; */
      }
   }
   
   data->currtype = type;

   func = &(data->player[data->currtype]);

   if( func->player_init == NULL )
      return 0;
   
   r = func->player_init( func->data );
   
   return r;
}

gint player_close( PlayerData *data )
{
   PlayerInterface *func;
   gint r;
   
   func = &(data->player[data->currtype]);
   
   if( func->player_close == NULL )
      return 0;
   
   r = func->player_close( func->data );
   
   return r;
}

gint player_play( PlayerData *data, player_type type, const gchar *filename )
{
   PlayerInterface *func;
   gint r;
   
   if( player_init(data, type) < 0 )
      return -1;
   
   func = &(data->player[data->currtype]);
   
   if( func->player_play == NULL )
      return -1;
   
   r = func->player_play( func->data, filename );
   data->ispaused = FALSE;
   
   return r;
}

gint player_stop( PlayerData *data )
{
   PlayerInterface *func;
   gint r;
   
   func = &(data->player[data->currtype]);
   
   if( func->player_stop == NULL )
      return -1;
   
   /* if( data->ispaused )
   {  if( player_toggle_pause(data) < 0 )
         return -1;
   } */
   
   r = func->player_stop( func->data );
   
   return r;
}

gint player_toggle_pause( PlayerData *data )
{
   PlayerInterface *func;
   gint r;
   
   func = &(data->player[data->currtype]);
   
   if( func->player_suspend == NULL || func->player_continue == NULL )
      return 0;
   
   if(!data->ispaused)
      r = func->player_suspend( func->data );
   else
      r = func->player_continue( func->data );
   
   data->ispaused = !data->ispaused;
   
   return r;
}

gint player_seek( PlayerData *data, gdouble seconds )
{
   PlayerInterface *func;
   gint r;
   
   func = &(data->player[data->currtype]);
   
   if( func->player_seek == NULL )
      return 0;
   
   r = func->player_seek( func->data, seconds );
   
   return r;
}

gint player_get_progress( PlayerData *data, gdouble *seconds_elap, gdouble *seconds_rem )
{
   PlayerInterface *func;
   gint r;
   
   func = &(data->player[data->currtype]);
   
   if( func->player_progress == NULL )
   {
      /* *seconds_elap = -1;
      *seconds_rem = -1; */
      return -1;
   }
   
   r = func->player_progress( func->data, seconds_elap, seconds_rem );
   
   return r;
}

gchar* player_get_pattern( PlayerData *data, player_type type )
{
   return data->player[type].pattern;
}

void player_set_next_func( PlayerData *data, GetNextFunc nextfunc, gpointer user_data )
{
   data->get_next = nextfunc;
   data->next_data = user_data;
}

gboolean player_ispaused( PlayerData *data )
{
   return data->ispaused;
}
