/* playlist.c */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <regex.h>

#include "config.h"
#include "logger.h"
#include "playlist.h"

void playlist_free_e( playlist_element *ple )
{
   g_free(ple->descript);
   g_free(ple->absfilename);
   g_free(ple); /* is this necessary or handled by g_list_free()? */
}

void playlist_clear( PlayListData *data )
{
   g_list_foreach(data->list, (GFunc)playlist_free_e, NULL);
   g_list_free(data->list);
   g_list_free(data->rand);
   data->list = NULL;
   data->rand = NULL;
   data->lastacc = NULL;
   data->nlastacc = -1;
}

void playlist_free( PlayListData *data )
{
   playlist_clear(data);
   g_free(data);
}

gint playlist_sort_e( playlist_element *ple1, playlist_element *ple2 )
{
   return strcmp(ple1->descript, ple2->descript);
}

void playlist_sort( PlayListData *data, gint *pos )
{
   gpointer lpos;
   
   if( pos != NULL )
      lpos = g_list_nth_data(data->list, *pos);
   
   data->list = g_list_sort( data->list, (GCompareFunc)playlist_sort_e );
   
   if( pos != NULL )
      *pos = g_list_index(data->list, lpos);
}

gint playlist_shuffle_e( gpointer *p1, gpointer *p2 )
{
   return 1-(int)(3.0*rand()/(RAND_MAX+1.0));
}

void playlist_shuffle( PlayListData *data, gint *pos )
{
   gpointer lpos;
   
   if( pos != NULL )
      lpos = g_list_nth_data(data->list, *pos);
   
   data->list = g_list_sort( data->list, (GCompareFunc)playlist_shuffle_e );
   
   if( pos != NULL )
      *pos = g_list_index(data->list, lpos);
}

void playlist_randomize( PlayListData *data, gint n )
{
   gpointer elem;
   
   data->rand = g_list_sort( data->rand, (GCompareFunc)playlist_shuffle_e );
   
   if( n >= 0 )
   {
      elem = g_list_nth_data(data->list, n);
      if( elem == NULL ) return;
      data->rand = g_list_remove(data->rand, elem);
      data->rand = g_list_prepend(data->rand, elem);
   }
}

void playlist_add( PlayListData *data, const char *dirname, const char *filename,
                   player_type player )
{
   playlist_element *ple;
   gint n;
   
   if( player <= PLAYER_NONE || player >= PLAYER_LAST )
      return; /* invalid player type */
   
   ple = g_new0(playlist_element, 1);
   ple->absfilename = g_strconcat(dirname, G_DIR_SEPARATOR_S, filename, NULL);
   ple->descript = g_strdup(filename);
   ple->ismarked = FALSE;
   ple->isplaying = FALSE;
   /* ple->noinfo = TRUE; */
   ple->player = player;
   
   data->list = g_list_append(data->list, ple);
   n = g_list_length(data->rand);
   n = (double)n*rand()/(RAND_MAX+1.0);
   data->rand = g_list_insert(data->rand, ple, n);
}

void playlist_delete( PlayListData *data, gint *pos )
{
   GList *link, *next;
   playlist_element *ple;
   gpointer lpos;
   
   if( pos != NULL )
      lpos = g_list_nth_data(data->list, *pos);
   
   link = g_list_first(data->list);
   while( link != NULL )
   {
      next = g_list_next(link);
      ple = (playlist_element*)link->data;
      if(ple->ismarked)
      {
         data->list = g_list_remove_link(data->list, link);
         data->rand = g_list_remove(data->rand, ple);
         playlist_free_e(ple);
      }
      link = next;
   }
   
   if( pos != NULL )
      *pos = g_list_index(data->list, lpos);
   
   data->lastacc = NULL;
   data->nlastacc = -1;
}

void playlist_move( PlayListData *data, guint *pos1, gint *pos2 )
{
   GList *link, *next, *stack=NULL;
   playlist_element *ple;
   gpointer lpos2;
   gint i, f;
   
   i=0; f=0;
   link = g_list_first(data->list);
   while( i<=*pos1 && link != NULL )
   {
      ple = (playlist_element*)link->data;
      if(ple->ismarked)
         f++;
      link = g_list_next(link);
      i++;
   }
   
   lpos2 = g_list_nth_data(data->list, *pos2);
   g_debug(__FILE__": got pointer %p for pos2", lpos2);
   
   link = g_list_first(data->list);
   while( link != NULL )
   {
      next = g_list_next(link);
      ple = (playlist_element*)link->data;
      if(ple->ismarked)
      {
         data->list = g_list_remove_link(data->list, link);
         stack = g_list_concat(stack, link);
      }
      link = next;
   }
   
   link = g_list_last(stack);
   while( link != NULL )
   {
      data->list = g_list_insert(data->list, link->data, *pos1-f+1);
      link = g_list_previous(link);
   }
   
   data->lastacc = NULL;
   data->nlastacc = -1;
   
   *pos1 = *pos1 - f;
   *pos2 = g_list_index(data->list, lpos2);
   
}

void playlist_print_e( playlist_element *ple, FILE *f)
{
   fprintf(f,"%s\n", ple->absfilename);
}

gint playlist_write( PlayListData *data, gchar *filename )
{
   FILE *f;
   
   f = fopen(filename, "w");
   if( f == NULL )
   {  g_info("Error opening file '%s': %s", filename, g_strerror(errno));
      return -1;
   }
   
   g_list_foreach(data->list, (GFunc)playlist_print_e, f);
   /* g_list_foreach(data->rand, (GFunc)playlist_print_e, f); */
   
   fclose(f);
   
   return 0;
}

playlist_element* playlist_get_element( PlayListData *data, guint n )
{
   /* same element as last time? */
   if(data->nlastacc >= 0 && data->nlastacc == n );
   /* do we only need the next/prev element of the cached? */
   else if( data->nlastacc >= 0 && data->nlastacc == n-1 )
      data->lastacc = g_list_next(data->lastacc);
   else if(data->nlastacc >= 0 && data->nlastacc == n+1 )
      data->lastacc = g_list_previous(data->lastacc);
   else
      data->lastacc = g_list_nth(data->list, n);
   
   if(data->lastacc == NULL) return NULL;
   
   data->nlastacc = n;
   
   return (playlist_element*)data->lastacc->data;
}

char* playlist_get_nth_name( PlayListData *data, guint n )
{
   playlist_element *ple;
   
   ple = playlist_get_element(data, n);
   
   if( ple == NULL )
      return NULL;
      
   return ple->descript;
}

char* playlist_get_nth_absname( PlayListData *data, guint n )
{
   playlist_element *ple;
   
   ple = playlist_get_element(data, n);
   
   if( ple == NULL )
      return NULL;
      
   return ple->absfilename;
}

player_type playlist_get_nth_type( PlayListData *data, guint n )
{
   playlist_element *ple;
   
   ple = playlist_get_element(data, n);
   
   if( ple == NULL )
      return PLAYER_NONE;
      
   return ple->player;
}

gint playlist_rand2i( PlayListData *data, guint *n )
{
   playlist_element *ple;
   
   ple = (playlist_element*)g_list_nth_data(data->rand, *n);
   
   if( ple == NULL ) return -1;
   
   *n = g_list_index(data->list, ple);
   
   return 0;
}

gint playlist_i2rand( PlayListData *data, guint *n )
{
   playlist_element *ple;
   
   ple = (playlist_element*)g_list_nth_data(data->list, *n);
   
   if( ple == NULL ) return -1;
   
   *n = g_list_index(data->rand, ple);
   
   return 0;
}

gboolean playlist_get_ismarked( PlayListData *data, guint n )
{
   playlist_element *ple;
   
   ple = playlist_get_element(data, n);
   
   if( ple == NULL )
      return FALSE;
   
   return ple->ismarked;
}

void playlist_set_ismarked( PlayListData *data, guint n, gboolean ismarked )
{
   playlist_element *ple;
   
   ple = playlist_get_element(data, n);
   
   if( ple == NULL )
      return;
   
   ple->ismarked = ismarked;
}

void playlist_regerror( const regex_t *preg, gint errnr )
{
   gchar *buf;
   size_t n;
   
   n = regerror(errnr, preg, NULL, 0);
   buf = g_malloc(n);
   regerror(errnr, preg, buf, n);
   
   g_info("regerror: %s", buf);
   g_free(buf);
}

void playlist_set_ismarked_regex( PlayListData *data, gboolean ismarked, const char *regex )
{
   GList *link;
   regex_t preg;
   playlist_element *ple;
   gint r;
   
   if( (r=regcomp(&preg, regex, REG_NOSUB|REG_EXTENDED)) != 0 )
   {  playlist_regerror(&preg, r);
      return;
   }
   
   link = g_list_first(data->list);
   while( link != NULL )
   {
      ple = (playlist_element*)link->data;
      if( regexec(&preg, ple->descript, 0, NULL, 0) == 0 )
         ple->ismarked = ismarked;
      
      link = g_list_next(link);
   }
   
   regfree(&preg);
}

guint playlist_get_length( PlayListData *data )
{
   return g_list_length(data->list);
}

PlayListData* playlist_new()
{
   PlayListData *data;
   
   data = g_new0(PlayListData, 1);
   data->list = NULL;
   data->rand = NULL;
   data->lastacc = NULL;
   data->nlastacc = -1;
   
   return data;
}
