/* screen.c */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/termios.h>

#include "config.h"
#include "logger.h"
#include "mcplay.h"
#include "filelist.h"
#include "playlist.h"
#include "screen.h"

/* declare private helper functions */
void screen_refresh( ScreenData *data );
void screen_filelist_print( ScreenData *data );
void screen_playlist_print( ScreenData *data );
void print_file_lines( ScreenData *data );
void print_play_lines( ScreenData *data );
void screen_set_progress( ScreenData *data, gdouble sec_pos, gdouble sec_rem );
gboolean screen_progress_timeout( ScreenData *data );


void screen_refresh( ScreenData *data )
{
   WINDOW *w;
   
   if(data->mode == SCREEN_FILE)
      w = data->fl;
   else if(data->mode == SCREEN_PLAY)
      w = data->pl;
   else if(data->mode == SCREEN_HELP)
      w = data->h;
   
   if( data->askpos > 0 )
      wmove(w, data->lines-1, data->askpos);
   
   wrefresh(w);
}

void screen_hardrefresh( ScreenData *data )
{
   gchar *dirname;
   
   wclear(data->fl);
   wattron(data->fl, A_BOLD);
   mvwaddstr(data->fl, 0, 0, "Filelist: ");
   dirname = filelist_get_dirname(data->filelist);
   mvwaddnstr(data->fl, 0, 10, dirname, data->cols-10);
   waddch(data->fl, '\n');
   wattroff(data->fl, A_BOLD);
   mvwhline(data->fl, 1, 0, ACS_HLINE, data->cols);
   mvwhline(data->fl, data->lines-2, 0, ACS_HLINE, data->cols);
   mvwaddstr(data->fl, data->lines-1, 0, data->strbg);
   print_file_lines(data);
   
   wclear(data->pl);
   wattron(data->pl, A_BOLD);
   mvwaddstr(data->pl, 0, 0, "Playlist");
   wattroff(data->pl, A_BOLD);
   if(data->repeat)
   {
      wattron(data->pl, A_BOLD);
      mvwaddstr(data->pl, 0, 10, "<repeat>");
      wattroff(data->pl, A_BOLD);
   }   
   if(data->random)
   {
      wattron(data->pl, A_BOLD);
      mvwaddstr(data->pl, 0, 20, "<random>");
      wattroff(data->pl, A_BOLD);
   }
   mvwhline(data->pl, 1, 0, ACS_HLINE, data->cols);
   mvwhline(data->pl, data->lines-2, 0, ACS_HLINE, data->cols);
   mvwaddstr(data->pl, data->lines-1, 0, data->strbg);
   print_play_lines(data);
   
   wclear(data->h);
   mvwaddstr(data->h, 0, 0, HELPTEXT);
   
   screen_refresh(data);
}

void screen_getsize( gint *lines, gint *cols )
{
   gint x, y;
   struct winsize win;
   
   ioctl(STDIN_FILENO, TIOCGWINSZ, &win);
   
   /* Avoid bogus sizes. */
   if (win.ws_col < 10 || win.ws_row < 5)
      return;
   
   /* we are brave here */
   LINES = win.ws_row;
   COLS = win.ws_col;
   
   /* Setup the ncurses library internal data. */
   resizeterm(win.ws_row, win.ws_col);
   g_debug(__FILE__": ioctl says: resize to %dx%d", win.ws_row, win.ws_col);
   *lines = win.ws_row;
   *cols = win.ws_col;
   
   /* endwin();
   initscr(); */
   clear();
   /* refresh(); */
   
   getmaxyx(stdscr, y, x);
   g_debug(__FILE__": getmaxyx says: resize to %dx%d", y, x);
   g_debug(__FILE__": LINES/COLS say: resize to %dx%d", LINES, COLS);
   
}

void screen_resize( ScreenData *data )
{         
   screen_getsize(&data->lines, &data->cols);
   
   mvaddch(data->lines-1, data->cols-1, '@');
   refresh();

   /* delwin(data->h);
   delwin(data->pb);
   delwin(data->pl);
   delwin(data->fb);
   delwin(data->fl);
   
   data->filepagesize = data->lines-4;
   data->fl = newwin(win.ws_row, win.ws_col, 0, 0);
   touchwin(data->fl);
   data->fb = subwin(data->fl, data->lines-4, data->cols, 2, 0);
   syncok(data->fb, TRUE);

   data->playpagesize = data->lines-4;
   data->pl = newwin(0, 0, 0, 0);
   touchwin(data->pl);
   data->pb = subwin(data->pl, data->lines-4, data->cols, 2, 0);
   syncok(data->pb, TRUE);
   
   data->h = newwin(0, 0, 0, 0); */
   
   /* wresize(data->fl, data->lines, data->cols); */
   /* wresize(data->fb, data->lines-2, data->cols); */
   /* wresize(data->pl, data->lines, data->cols); */
   /* wresize(data->pb, data->lines-2, data->cols); */
   /* wresize(data->h, data->lines, data->cols); */
   
   /* screen_hardrefresh(data); */
}

void screen_ask_start( ScreenData *data, const char *quest )
{
   WINDOW *w;
   
   if( data->mode == SCREEN_FILE )
      w = data->fl;
   else if( data->mode == SCREEN_PLAY )
      w = data->pl;
   else if( data->mode == SCREEN_HELP )
      w = data->h;
   else
   {  g_warning(__FILE__": Error in screen_ask_start: invalid mode");
      return;
   }
   
   mvwaddnstr(w, data->lines-1, 0, quest, data->cols);
   waddch(w, '\n');
   wrefresh(w);
   data->askpos = strlen(quest);
   data->nmsg = -1;
   curs_set(1);
   /* leaveok(w, FALSE); */
   echo();
   
   return;
}

gchar* screen_ask_getch( ScreenData *data )
{
   static gchar buf[INPUT_MAX];
   static gint i = 0;
   gint key;
   WINDOW *w;
   
   if( data->mode == SCREEN_FILE )
      w = data->fl;
   else if( data->mode == SCREEN_PLAY )
      w = data->pl;
   else if( data->mode == SCREEN_HELP )
      w = data->h;
   else
   {  g_warning(__FILE__": Error in screen_ask_getch: invalid mode");
      buf[0] = '\0';
      i = 0;
      curs_set(0);
      noecho();
      data->askpos = -1;
      data->nmsg = 0;
      screen_message(data, 0, data->strbg);
      return buf;
   }
   
   key = mvgetch(data->lines-1, data->askpos);
   
   if( key == '\033' || key == '\007' )
   {  buf[0] = '\0';
      i = 0;
      curs_set(0);
      noecho();
      data->askpos = -1;
      data->nmsg = 0;
      screen_message(data, 0, data->strbg);
      return buf;
   }
   else if( key == '\r' || key == KEY_ENTER )
   {  buf[i] = '\0';
      i = 0;
      curs_set(0);
      noecho();
      data->askpos = -1;
      data->nmsg = 0;
      screen_message(data, 0, data->strbg);
      return buf;
   }
   else if( key == KEY_BACKSPACE )
   {  i--;
      data->askpos--;
      if( i < 0 )
      {  i = 0;
         data->askpos++;
      }
      mvwaddch(w, data->lines-1, data->askpos, '\n');
      wrefresh(w);
      return NULL;
   }
   else if( key >= 0x20 && key < KEY_MIN ) /* an 'ordinary' key */
   {  buf[i] = (gchar)key;
      i++;
      data->askpos++;
      if( i >= sizeof(buf) )
      {  i = sizeof(buf)-1;
         data->askpos--;
      }
      return NULL;
   }
   else
   {  fprintf(stderr, "\a");
      fflush(stderr);
   }

   
   /* mvwaddch(w, data->lines-1, 0, '\n'); */
   /* screen_message(data, 0, data->strbg); */
   
   return NULL;
}

gboolean screen_timeout( ScreenData *data )
{
   data->nmsg--;
   if( data->nmsg == 0 )
      screen_message(data, -1, data->strbg);
      
   return FALSE;
}

void screen_message( ScreenData *data, gint timeout, const gchar *fmt, ... )
{
   va_list az;
   gchar *tmp, *str;
   
   va_start(az, fmt);
   str = g_strdup_vprintf(fmt, az);
   va_end(az);
   
   if( data->nmsg >= 0 )
   {
      tmp = g_strdup_printf("%*.*s", -1*(data->cols-7), data->cols-7, str);
      mvwaddstr(data->fl, data->lines-1, 0, tmp);
      mvwaddstr(data->pl, data->lines-1, 0, tmp);
      g_free(tmp);
      screen_refresh(data);
   }
   
   
   if(timeout > 0)
   {
      g_free(str);
      data->nmsg++;
      g_timeout_add(timeout*1000, (GSourceFunc)screen_timeout, data);
   }
   else
   {
      g_free(data->strbg);
      data->strbg = str;
   }
}

void print_file_lines( ScreenData *data )
{
   WINDOW *w;
   guint nlines, start, num;
   gint selected;
   gint i;
   
   w = data->fb;
   nlines = data->nfilestr;
   start = data->filesel/data->filepagesize*data->filepagesize;
   num = data->filepagesize;
   selected = data->filesel;
    
   for(i=0; i<num; i++)
   {
      wmove(w, i, 0);
      if( i+start == selected ) wattron(w, A_REVERSE);
      if( i+start < nlines )
         waddnstr(w, filelist_get_nth_name(data->filelist,i+start), data->cols);
      waddch(w, '\n');
      if( i+start == selected ) wattroff(w, A_REVERSE);
   }
}

void print_play_lines( ScreenData *data )
{
   WINDOW *w;
   guint nlines, start, num;
   gint selected, highlight;
   gint i;
   
   w = data->pb;
   nlines = data->nplaystr;
   start = data->playsel/data->playpagesize*data->playpagesize;
   num = data->playpagesize;
   selected = data->playsel;
   highlight = data->nplaycurr;
   if( data->random )
      playlist_rand2i(data->playlist, &highlight);
   
   for(i=0; i<num; i++)
   {
      wmove(w, i, 0);
      if( i+start < nlines )
      {
         if( i+start == selected )
            wattron(w, A_REVERSE);
         if( i+start == highlight )
            wattron(w, A_BOLD);
         if( playlist_get_ismarked(data->playlist, i+start) )
         {  if( useColors )
               wattron(w, COLOR_PAIR(COLORMARK));
            else
               waddstr(w, "# ");
         }
         else
         {  if( !useColors )
               waddstr(w, "  ");
         }
         
         waddnstr(w, playlist_get_nth_name(data->playlist,i+start), data->cols);
         
         if( i+start == selected )
            wattroff(w, A_REVERSE);
         if( i+start == highlight )
            wattroff(w, A_BOLD);
         if( playlist_get_ismarked(data->playlist, i+start) )
            wattroff(w, COLOR_PAIR(COLORMARK));
      }
      waddch(w, '\n');
   }
}

void screen_filelist_print( ScreenData *data )
{
   gchar *dirname;
   
   dirname = filelist_get_dirname(data->filelist);
   wattron(data->fl, A_BOLD);
   mvwaddnstr(data->fl, 0, 10, dirname, data->cols-10);
   waddch(data->fl, '\n');
   wattroff(data->fl, A_BOLD);
   
   print_file_lines(data);
   
   if( data->mode == SCREEN_FILE )
      wrefresh(data->fl);
}

gboolean screen_progress_timeout( ScreenData *data )
{
   gdouble secpos, secrem;
   gint r;
   
   r = player_get_progress(data->player, &secpos, &secrem);
   /* g_debug(__FILE__": got %g %g (%d)", secpos, secrem, r); */
   
   if( r < 0 )
   {  screen_set_progress(data, -1, -1);
      return FALSE;
   }
   
   screen_set_progress(data, secpos, secrem);
   
   return TRUE;
}

void screen_filelist_up( ScreenData *data )
{
   data->filesel--;
   if( data->filesel < 0 ) data->filesel = data->nfilestr-1;
   
   screen_filelist_print(data);
}

void screen_filelist_down( ScreenData *data )
{
   data->filesel++;
   if( data->filesel >= data->nfilestr ) data->filesel = 0;

   screen_filelist_print(data);
}

void screen_filelist_pageup( ScreenData *data )
{
   data->filesel = data->filesel - data->filepagesize - data->filesel%data->filepagesize;
   if( data->filesel < 0 ) data->filesel = (data->nfilestr-1)/data->filepagesize*data->filepagesize;
   
   screen_filelist_print(data);
}

void screen_filelist_pagedown( ScreenData *data )
{
   data->filesel = data->filesel + data->filepagesize - data->filesel%data->filepagesize;
   if( data->filesel >= data->nfilestr ) data->filesel = 0;

   screen_filelist_print(data);
}

void screen_filelist_home( ScreenData *data )
{
   data->filesel = 0;
   
   screen_filelist_print(data);
}

void screen_filelist_end( ScreenData *data )
{
   data->filesel = data->nfilestr-1;
   
   screen_filelist_print(data);
}


void screen_playlist_print( ScreenData *data )
{
   if(data->repeat)
   {
      wattron(data->pl, A_BOLD);
      mvwaddstr(data->pl, 0, 10, "<repeat>");
      wattroff(data->pl, A_BOLD);
   }
   else
      mvwaddstr(data->pl, 0, 10, "        ");
   
   if(data->random)
   {
      wattron(data->pl, A_BOLD);
      mvwaddstr(data->pl, 0, 20, "<random>");
      wattroff(data->pl, A_BOLD);
   }
   else
      mvwaddstr(data->pl, 0, 20, "        ");
   
   print_play_lines(data);
   
   if( data->mode == SCREEN_PLAY )
      wrefresh(data->pl);
}

void screen_playlist_up( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   data->playsel--;
   if( data->playsel < 0 ) data->playsel = data->nplaystr-1;
   
   screen_playlist_print(data);
}

void screen_playlist_down( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   data->playsel++;
   if( data->playsel >= data->nplaystr ) data->playsel = 0;

   screen_playlist_print(data);
}

void screen_playlist_pageup( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   data->playsel = data->playsel - data->playpagesize - data->playsel%data->playpagesize;
   if( data->playsel < 0 ) data->playsel = (data->nplaystr-1)/data->playpagesize*data->playpagesize;
   
   screen_playlist_print(data);
}

void screen_playlist_pagedown( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   data->playsel = data->playsel + data->playpagesize - data->playsel%data->playpagesize;
   if( data->playsel >= data->nplaystr ) data->playsel = 0;

   screen_playlist_print(data);
}

void screen_playlist_home( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   data->playsel = 0;
   
   screen_playlist_print(data);
}

void screen_playlist_end( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   data->playsel = data->nplaystr-1;
   
   screen_playlist_print(data);
}

void screen_filelist_select( ScreenData *data )
{
   gchar *name;
   
   name = filelist_get_nth_absname(data->filelist, data->filesel);
   
   if(filelist_nth_isdir(data->filelist, data->filesel))
   {
      g_debug(__FILE__": selected dir");
      screen_filelist_chng_dir(data, name);
   }
   else if(filelist_nth_isfile(data->filelist, data->filesel))
   {
      g_debug(__FILE__": selected file");
      data->nplaycurr = -1;
      player_play(data->player,
                  filelist_get_nth_type(data->filelist, data->filesel),
                  name);
      screen_message(data, 0, "Playing: %s", filelist_get_nth_name(data->filelist, data->filesel));
      g_timeout_add(200, (GSourceFunc)screen_progress_timeout, data);
   }
   
   g_free(name);
}

gint screen_filelist_add_dir( ScreenData *data, gchar *dirname )
{
   gint i, n;
   gchar *odir;
   gchar *rdir;
   
   /* we save the dirname here to restore it after adding all
      the files in the dir. changing to '..' afterwards
      will not work on linked dirs! */
   odir = g_strdup(filelist_get_dirname(data->filelist));
   
   if( filelist_open_dir(data->filelist, dirname) < 0 )
   {  g_info("Error adding dir '%s' recursively: %s", dirname, g_strerror(errno));
      return -1;
   }
   
   n = filelist_get_length(data->filelist);
   
   /* parent dir has index 0 so we start from index 1 */
   for(i=1; i<n; i++)
   {
      if( filelist_nth_isfile(data->filelist, i) )
         playlist_add(data->playlist,
                      filelist_get_dirname(data->filelist),
                      filelist_get_nth_name(data->filelist, i),
                      filelist_get_nth_type(data->filelist, i));
      else if( filelist_nth_isdir(data->filelist, i) )
      {
         rdir = g_strdup(filelist_get_nth_name(data->filelist, i));
         screen_filelist_add_dir(data, rdir);
         g_free(rdir);
      }
   }
   
   if( filelist_open_dir(data->filelist, odir) < 0 )
   {  g_info("Error in leaving dir: %s", g_strerror(errno));
      return -1;
   }
   g_free(odir);
   
   data->nplaystr = playlist_get_length(data->playlist);
   if( data->playsel < 0 && data->nplaystr > 0 ) data->playsel = 0;
   
   return 0;
}

void screen_filelist_mark( ScreenData *data )
{
   gchar *name;
   
   if( filelist_nth_isfile(data->filelist, data->filesel) )
   {
      g_debug(__FILE__": marked file");
      playlist_add(data->playlist,
                   filelist_get_dirname(data->filelist),
                   filelist_get_nth_name(data->filelist, data->filesel),
                   filelist_get_nth_type(data->filelist, data->filesel));
   }
   else if( filelist_nth_isdir(data->filelist, data->filesel) )
   {
      g_debug(__FILE__": marked dir");
      /* name = g_strdup(filelist_get_dirname(data->filelist)); */
      filelist_add_dir_r(data->filelist,
                         filelist_get_nth_name(data->filelist, data->filesel),
                         (PlaylistAddFunc)playlist_add,
                         data->playlist );
      /* filelist_open_dir(data->filelist, name); */
      /* g_free(name); */
   }
   else if( filelist_nth_islist(data->filelist, data->filesel) )
   {
      g_debug(__FILE__": marked playlist");
      name = filelist_get_nth_absname(data->filelist, data->filesel);
      filelist_open_list(data->filelist, name, (PlaylistAddFunc)playlist_add, data->playlist);
      g_free(name);
   }
   
   data->nplaystr = playlist_get_length(data->playlist);
   if( data->playsel < 0 && data->nplaystr > 0 ) data->playsel = 0;
   
   screen_filelist_down(data);
}

void screen_filelist_chng_dir( ScreenData *data, const char *dirname )
{
   gint i;
   
   filelist_open_dir(data->filelist, dirname);
   i=filelist_get_olddir(data->filelist);
   data->nfilestr = filelist_get_length(data->filelist);
   if(i>=0) data->filesel = i;
   else data->filesel = 0;
   screen_filelist_print(data);
}

void screen_playlist_select( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   player_play(data->player,
               playlist_get_nth_type(data->playlist, data->playsel),
               playlist_get_nth_absname(data->playlist, data->playsel));
   
   if( data->random )
   {
      playlist_randomize(data->playlist, data->playsel);
      data->nplaycurr = 0;
   }
   else
      data->nplaycurr = data->playsel;
   
   screen_message(data, 0, "Playing: %s", playlist_get_nth_name(data->playlist, data->playsel));
   g_timeout_add(200, (GSourceFunc)screen_progress_timeout, data);
   
   screen_playlist_print(data);
}

void screen_playlist_mark( ScreenData *data )
{
   gboolean ismarked;

   if( data->playsel < 0 ) return;   
   
   ismarked = playlist_get_ismarked(data->playlist, data->playsel);
   playlist_set_ismarked(data->playlist, data->playsel, !ismarked);
   
   screen_playlist_down(data);
   screen_playlist_print(data);
}

void screen_playlist_mark_all( ScreenData *data )
{
   playlist_set_ismarked_regex(data->playlist, TRUE, ".*");
   screen_playlist_print(data);
}

void screen_playlist_unmark_all( ScreenData *data )
{
   playlist_set_ismarked_regex(data->playlist, FALSE, ".*");
   screen_playlist_print(data);
}

void screen_playlist_mark_reg( ScreenData *data, const gchar *regex )
{
   playlist_set_ismarked_regex(data->playlist, TRUE, regex);
   screen_playlist_print(data);
}

void screen_playlist_unmark_reg( ScreenData *data, const gchar *regex )
{
   playlist_set_ismarked_regex(data->playlist, FALSE, regex);
   screen_playlist_print(data);
}

void screen_playlist_delete( ScreenData *data )
{
   if( data->playsel < 0 ) return;
   
   playlist_delete( data->playlist, &data->nplaycurr );
   data->nplaystr = playlist_get_length(data->playlist);
   if(data->playsel > data->nplaystr-1)
      data->playsel = data->nplaystr-1;
   
   screen_playlist_print(data);
}

void screen_playlist_move( ScreenData *data )
{
   if( data->playsel < 0 ) return;

   if( data->random )
      playlist_rand2i(data->playlist, &data->nplaycurr);
   
   playlist_move(data->playlist, &data->playsel, &data->nplaycurr);
   /* data->playsel = i; */
   
   if( data->random )
      playlist_i2rand(data->playlist, &data->nplaycurr);
   
   screen_playlist_print(data);
}

void screen_playlist_shuffle( ScreenData *data )
{
   if( data->random )
      playlist_rand2i(data->playlist, &data->nplaycurr);
   
   playlist_shuffle(data->playlist, &data->nplaycurr);
   
   if( data->random )
      playlist_i2rand(data->playlist, &data->nplaycurr);
   
   screen_playlist_print(data);
}

void screen_playlist_sort( ScreenData *data )
{
   if( data->random )
      playlist_rand2i(data->playlist, &data->nplaycurr);
   
   playlist_sort(data->playlist, &data->nplaycurr);
   
   if( data->random )
      playlist_i2rand(data->playlist, &data->nplaycurr);
   
   screen_playlist_print(data);
}

void screen_playlist_toggle_repeat( ScreenData *data )
{
   data->repeat = !data->repeat;
   screen_playlist_print(data);
}

void screen_playlist_toggle_random( ScreenData *data )
{
   data->random = !data->random;
   
   if( data->random && data->nplaycurr >= 0 )
   {
      playlist_randomize(data->playlist, data->nplaycurr);
      data->nplaycurr = 0;
   }
   if( !data->random && data->nplaycurr >= 0 )
   {
      playlist_rand2i(data->playlist, &data->nplaycurr);
   }
   
   screen_playlist_print(data);
}

gint screen_playlist_add_unknown( ScreenData *data, const gchar *filename )
{
   gint r;
   
   r = filelist_open_unknown(data->filelist, filename, (PlaylistAddFunc)playlist_add, data->playlist);
         
   data->nplaystr = playlist_get_length(data->playlist);
   if( data->playsel < 0 && data->nplaystr > 0 ) data->playsel = 0;
   
   return r;
}

void screen_playlist_write( ScreenData *data, gchar *filename )
{
   playlist_write(data->playlist, filename);
}

void screen_player_stop( ScreenData *data )
{
   screen_message(data, 0, "");
   
   player_stop(data->player);
   
   data->nplaycurr = -1;
   
   screen_playlist_print(data);
}

void screen_player_next( ScreenData *data )
{
   gint n;
      
   if( data->nplaystr <= 0 || data->nplaycurr < 0 )
   {  screen_message(data, 0, "");
      return;
   }
   
   data->nplaycurr++;
   if( data->nplaycurr >= data->nplaystr )
   {
      if( !data->repeat )
      {  /* player_stop(data->player);
         data->nplaycurr = -1;
         screen_playlist_print(data); */
         data->nplaycurr = data->nplaystr-1;
         return;
      }
      else
         data->nplaycurr = 0;
   }
   
   n = data->nplaycurr;
   if( data->random )
      playlist_rand2i(data->playlist, &n);
      
   player_play(data->player,
               playlist_get_nth_type(data->playlist, n),
               playlist_get_nth_absname(data->playlist, n));
   
   screen_message(data, 0, "Playing: %s", playlist_get_nth_name(data->playlist, n));
   g_timeout_add(200, (GSourceFunc)screen_progress_timeout, data);
   
   screen_playlist_print(data);
}

void screen_player_prev( ScreenData *data )
{
   gint n;
   
   if( data->nplaystr <= 0 || data->nplaycurr < 0 )
   {  screen_message(data, 0, "");
      return;
   }
   
   data->nplaycurr--;
   if( data->nplaycurr < 0 )
   {
      if( !data->repeat )
      {  /* player_stop(data->player);
         data->nplaycurr = -1;
         screen_playlist_print(data); */
         data->nplaycurr = 0;
         return;
      }
      else
         data->nplaycurr = data->nplaystr-1;
   }
   
   n = data->nplaycurr;
   if( data->random )
      playlist_rand2i(data->playlist, &n);
   
   player_play(data->player,
               playlist_get_nth_type(data->playlist, n),
               playlist_get_nth_absname(data->playlist, n));
   
   screen_message(data, 0, "Playing: %s", playlist_get_nth_name(data->playlist, n));
   g_timeout_add(200, (GSourceFunc)screen_progress_timeout, data);
   
   screen_playlist_print(data);
}

void screen_player_toggle_pause( ScreenData *data )
{
   gchar *song;
   
   if( player_toggle_pause(data->player) < 0 )
      return;
   
   song = strchr(data->strbg, ' ');
   
   if( song == NULL )
   {  g_warning(__FILE__": screen status line contains bad data");
      return;
   }
   
   if( player_ispaused(data->player) )
      screen_message(data, 0, "Paused: %s", song+1);
   else
      screen_message(data, 0, "Playing: %s", song+1);
}

void screen_player_seek( ScreenData *data, gdouble seconds )
{
   player_seek(data->player, seconds);
}

void screen_toggle_timer( ScreenData *data )
{
   data->timer++;
   
   if( data->timer == TIMER_LAST )
      data->timer = TIMER_FWD;
}

void screen_set_progress( ScreenData *data, gdouble sec_pos, gdouble sec_rem )
{
   guint width;
   timer_mode timer;
   chtype c, o;
   
   if( sec_pos < 0 || sec_rem < 0 )
      width = 0;
   else
      width = 1 + data->cols*sec_pos/(sec_pos+sec_rem+0.1);
   
   if( useColors )
   {  c = ACS_HLINE|COLOR_PAIR(COLORPROG)|A_BOLD;
      o = ACS_HLINE|COLOR_PAIR(COLORPROG)|A_BOLD;
   }
   else
   {  c = '=';
      o = 'O';
   }
   
   mvwhline(data->pl, data->lines-2, 0, ACS_HLINE, data->cols);
   whline(data->pl, c, width-1);
   mvwaddch(data->pl, data->lines-2, width-1, o);
   
   mvwhline(data->fl, data->lines-2, 0, ACS_HLINE, data->cols);
   whline(data->fl, c, width-1);
   mvwaddch(data->fl, data->lines-2, width-1, o);
   
   if( sec_pos < 0 || sec_rem < 0 )
      timer = TIMER_NONE;
   else
      timer = data->timer;
   
   switch(timer)
   {
      case TIMER_NONE:
         mvwaddch(data->fl, data->lines-1, data->cols-7, '\n');
         mvwaddch(data->pl, data->lines-1, data->cols-7, '\n');
         break;
      case TIMER_FWD:
         mvwprintw(data->fl, data->lines-1, data->cols-7, "[%02d:%02d]", (gint)sec_pos/60, (gint)sec_pos%60);
         mvwprintw(data->pl, data->lines-1, data->cols-7, "[%02d:%02d]", (gint)sec_pos/60, (gint)sec_pos%60);
         break;
      case TIMER_BWD:
         mvwprintw(data->fl, data->lines-1, data->cols-7, "[%02d:%02d]", (gint)sec_rem/60, (gint)sec_rem%60);
         mvwprintw(data->pl, data->lines-1, data->cols-7, "[%02d:%02d]", (gint)sec_rem/60, (gint)sec_rem%60);
         break;
      default:
         g_critical(__FILE__": Error: unknown timer mode");
   }
   
   screen_refresh(data);
}

void screen_set_mode( ScreenData *data, screen_mode mode )
{
   data->mode = mode;
   
   if(mode == SCREEN_FILE)
   {
      redrawwin(data->fl);
      screen_filelist_print(data);
   }
   else if(mode == SCREEN_PLAY)
   {
      redrawwin(data->pl);
      screen_playlist_print(data);
   }
   else if(mode == SCREEN_HELP)
   {
      redrawwin(data->h);
      wrefresh(data->h);
   }
   else
   {
      g_critical(__FILE__": Error: unknown screen mode requested");
   }
}

screen_mode screen_get_mode( ScreenData *data )
{
   return data->mode;
}

ScreenData* screen_new()
{
   ScreenData *data;
   WINDOW *fl, *fb, *pl, *pb, *h;
   /* gint lines, cols; */
   
   initscr();      /* initialize the curses library */
   start_color();
   use_default_colors();
   keypad(stdscr, TRUE);  /* enable keyboard mapping */
   nonl();         /* tell curses not to do NL->CR/NL on output */
   cbreak();       /* take input chars one at a time, no wait for \n */
   noecho();         /* don't echo input */
   curs_set(0);     /* set cursor invisible */
   nodelay(stdscr, TRUE); /* return from getch() without blocking */
   g_atexit((void*)endwin); /* auf jeden fall beim beenden curses abschalten */
   
   if(has_colors())
   {  init_pair(COLORPROG, COLOR_CYAN, -1);
      init_pair(COLORMARK, COLOR_GREEN, -1);
   }
   else useColors = FALSE;
   
   refresh();
   /* resizeterm(lines, cols); */
   /* screen_getsize(&lines, &cols); */
     
   data = g_malloc0(sizeof(ScreenData));
   data->lines = LINES; /* LINES */
   data->cols = COLS; /* COLS */
   data->mode = SCREEN_FILE;
   data->repeat = FALSE;
   data->random = FALSE;
   data->nmsg = 0;
   data->strbg = NULL;
   data->timer = TIMER_BWD;
   data->askpos = -1;
   
   data->filepagesize = data->lines-4;
   data->nfilestr = 0;
   data->filesel = 0;
   
   fl = newwin(0, 0, 0, 0);
   touchwin(fl);
   fb = subwin(fl, data->filepagesize, data->cols, 2, 0);
   syncok(fb, TRUE);	/* sync changes on subwindow to window */
   
   data->playpagesize = data->lines-4;
   data->nplaystr = 0;
   data->playsel = -1;
   data->nplaycurr = -1;
   
   pl = newwin(0, 0, 0, 0);
   touchwin(pl);
   pb = subwin(pl, data->playpagesize, data->cols, 2, 0);
   syncok(pb, TRUE);
   
   h = newwin(0, 0, 0, 0);
   
   data->fl = fl;
   data->fb = fb;
   data->pl = pl;
   data->pb = pb;
   data->h = h;
   
   data->filelist = filelist_new();
      
   filelist_open_dir(data->filelist, ".");
   data->nfilestr = filelist_get_length(data->filelist);
   
   data->playlist = playlist_new();
   data->nplaystr = playlist_get_length(data->playlist);
   
   data->player = NULL;
   
   screen_hardrefresh(data);
   
   return data;
}

void screen_set_player( ScreenData *data, PlayerData *player )
{
   gint i;
   
   for(i=0; i<PLAYER_LAST; i++)
      filelist_register_pattern(data->filelist, i, player_get_pattern(player, i));
   
   player_set_next_func(player, (GetNextFunc)screen_player_next, data);
   
   /* reread dir with new patterns */
   screen_filelist_chng_dir(data, ".");
   
   data->player = player;
}

/* EOF */
